# encoding: utf-8
import tkinter as tk
import tkinter.ttk as ttk
from tkinter import filedialog


class PathChooserBase:
    """Allows to choose a file or directory.

    Encapsulates the usage of the functions:

        filedialog.askopenfilename()
        filedialog.asksaveasfilename(**fdoptions)
        filedialog.askdirectory()

    Generates <<PathChooserPathChanged>> event when the path is changed.

    Dialog options:
      initialdir: str
      filetypes: iterable
      title: str
      mustexist: bool
      defaultextension: str
    """

    FILE = "file"
    DIR = "directory"

    def __init__(self, *args, **kw):
        self._choose = self.FILE
        self._oldvalue = None
        self._path = None
        self._fdoptions = {
            "filetypes": tuple(),
            "initialdir": None,
            "mustexist": True,
            "title": None,
            "defaultextension": None,
        }
        super().__init__(*args, **kw)

    def configure(self, cnf=None, **kw):
        if cnf:
            return super().configure(cnf, **kw)
        key = "type"
        if key in kw:
            self._choose = kw.pop(key)
        key = "path"
        if key in kw:
            self._path = kw.pop(key)
        # dialog options
        for key in tuple(kw.keys()):
            if key in self._fdoptions:
                self._fdoptions[key] = kw.pop(key)
        return super().configure(cnf, **kw)

    config = configure

    def cget(self, key):
        option = "type"
        if key == option:
            return self._choose
        option = "path"
        if key == option:
            return self._path
        # dialog options
        if key in self._fdoptions:
            return self._fdoptions[key]
        return super().cget(key)

    __getitem__ = cget

    def _is_changed(self):
        if self._oldvalue != self._get_user_selected_path():
            return True
        return False

    def _get_user_selected_path(self):
        return self._path

    def _generate_changed_event(self):
        if self._is_changed():
            self._oldvalue = self._get_user_selected_path()
            self.event_generate("<<PathChooserPathChanged>>")

    def _do_path_selection(self):
        fname = None
        fdoptions = self._fdoptions.copy()
        fdoptions["parent"] = self.winfo_toplevel()
        if fdoptions["initialdir"] is None:
            fdoptions["initialdir"] = self.cget("path")
        if self._choose == self.FILE:
            must_exists = fdoptions.pop("mustexist")
            if must_exists:
                fname = filedialog.askopenfilename(**fdoptions)
            else:
                fname = filedialog.asksaveasfilename(**fdoptions)
        else:
            fdoptions.pop("filetypes")
            fdoptions.pop("defaultextension")
            fname = filedialog.askdirectory(**fdoptions)
        if fname:
            self.configure(path=fname)
            self._generate_changed_event()


class PathChooserInput(PathChooserBase, ttk.Frame):
    """Allows to choose a file or directory.

    Encapsulates the usage of the functions:

        filedialog.askopenfilename()
        filedialog.asksaveasfilename(**fdoptions)
        filedialog.askdirectory()

    Generates <<PathChooserPathChanged>> event when the path is changed.

    Dialog options:
      initialdir: str
      filetypes: iterable
      title: str
      mustexist: bool
      defaultextension: str

    Usage Example:
        # Choose filename for open:
        pcifile = PathChooserInput(framex)
        pcifile.config(
            initialdir='/home',
            title='Open file:',
            type='file',
            filetypes=[('text files-', '.txt'), ('uifiles', '.ui')]
            )
        pcifile.pack(fill='x', side='top')

        # Choose filename for save:
        pcofile = PathChooserInput(framex)
        pcofile.config(
            initialdir='/home',
            title='Save to:',
            type='file',
            mustexist=False,
            defaultextension=".txt",
            )
        pcofile.pack(fill='x', side='top')

        # Choose directory:
        pcidir = PathChooserInput(framex)
        pcidir.config(initialdir='/usr/local', mustexist=True,
                      title='Choose a directory:', type='directory')
        pcidir.pack(fill='x', side='top')
    """

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._state = "normal"
        # subwidgets
        self.entry = o = ttk.Entry(self, state=self._state)
        o.grid(row=0, column=0, sticky="ew")
        o.bind("<KeyPress>", self.__on_enter_key_pressed)
        o.bind("<FocusOut>", self.__on_focus_out)
        self.folder_button = o = ttk.Button(
            self,
            text="▶",
            command=self._do_path_selection,
            width=4,
            state=self._state,
        )
        o.grid(row=0, column=1, padx=2)

        # self.rowconfigure(0, weight = 0)
        self.columnconfigure(0, weight=1)

    def configure(self, cnf=None, **kw):
        if cnf:
            return super().configure(cnf, **kw)
        key = "image"
        if key in kw:
            self.folder_button.configure(image=kw.pop(key))
        key = "path"
        if key in kw:
            self.entry.delete(0, "end")
            self.entry.insert(0, kw.pop(key))
            self._generate_changed_event()
        key = "textvariable"
        if key in kw:
            self.entry.configure(textvariable=kw.pop(key))
            self._generate_changed_event()
        key = "state"
        if key in kw:
            value = kw.pop(key)
            self.entry.config(state=value)
            if value in ("disabled", "readonly"):
                self.folder_button.config(state="disabled")
            else:
                self.folder_button.config(state=value)
        return super().configure(cnf, **kw)

    config = configure

    def cget(self, key):
        option = "image"
        if key == option:
            return self.folder_button.cget(key)
        option = "path"
        if key == option:
            return self.entry.get()
        option = "textvariable"
        if key == option:
            return self.entry.cget(option)
        option = "state"
        if key == option:
            return self.entry.cget(option)
        return super().cget(key)

    __getitem__ = cget

    def _get_user_selected_path(self):
        return self.entry.get()

    def __on_enter_key_pressed(self, event):
        key = event.keysym
        if key in ("Return", "KP_Enter"):
            self._generate_changed_event()

    def __on_focus_out(self, event):
        self._generate_changed_event()


class PathChooserButton(PathChooserBase, ttk.Button):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.configure(command=self._do_path_selection)



import json


class ListDTO:
    def transform(self, value: str):
        val = None
        try:
            val = json.loads(value)
        except json.JSONDecodeError:
            val = ["invalid data"]
        return val



from pygubu.i18n import _
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from tkintermapview import TkinterMapView
from ..tkintermapview import _designer_tab_label, _plugin_uid


class TkinterMapViewBuilder(BuilderObject):
    class_ = TkinterMapView
    _int_props = ("width", "height", "corner_radius", "max_zoom")
    OPTIONS_CUSTOM = _int_props + ("bg_color",)
    ro_properties = OPTIONS_CUSTOM

    def _process_property_value(self, pname, value):
        if pname in self._int_props:
            return int(value)
        return super()._process_property_value(pname, value)

    def _code_process_property_value(self, targetid, pname, value: str):
        if pname in self._int_props:
            return self._process_property_value(pname, value)
        return super()._code_process_property_value(targetid, pname, value)


_builder_uid = f"{_plugin_uid}.TkinterMapView"
register_widget(
    _builder_uid,
    TkinterMapViewBuilder,
    "TkinterMapView",
    ("ttk", _designer_tab_label),
)

_none = {}
# pname, editor, options
_properties = (
    ("widht", "naturalnumber", _none),
    ("height", "naturalnumber", _none),
    (
        "corner_radius",
        "choice",
        {"values": [""] + [x for x in range(0, 31)], "state": "readonly"},
    ),
    ("bg_color", "colorentry", _none),
    (
        "max_zoom",
        "choice",
        {"values": [x for x in range(1, 20)], "state": "readonly"},
    ),
)

for pname, editor, options in _properties:
    register_custom_property(_builder_uid, pname, editor, **options)



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk


import fixpath
import pygubu
import support


class TestTkSpinbox(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_tkspinbox.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        self.widget = self.builder.get_object("test_from")
        self.spinbox = self.builder.get_object("spinbox1")
        self.assertIsInstance(self.spinbox, tk.Spinbox)
        self.widget.destroy()

    def test_from_(self):
        self.widget = self.builder.get_object("test_from")
        spinbox = self.builder.get_object("spinbox1")
        value = spinbox.cget("from")
        self.assertEqual(5, value)
        self.widget.destroy()

    def test_to(self):
        self.widget = self.builder.get_object("test_to")
        spinbox = self.builder.get_object("spinbox2")
        value = spinbox.cget("to")
        self.assertEqual(10, value)
        self.widget.destroy()

    def test_from_to(self):
        self.widget = self.builder.get_object("test_from_to")
        spinbox = self.builder.get_object("spinbox3")
        value_from = spinbox.cget("from")
        value_to = spinbox.cget("to")
        self.assertEqual(2, value_from)
        self.assertEqual(10, value_to)
        self.widget.destroy()



import awesometkinter as atk
from pygubu.i18n import _
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.tk.tkstdwidgets import TKCheckbutton
from pygubu.plugins.ttk.ttkstdwidgets import TTKButton, TTKRadiobutton
from ..awesometkinter import _designer_tab_label, _plugin_uid


class Button3dBO(TTKButton):
    OPTIONS_STANDARD = tuple(set(TTKButton.OPTIONS_STANDARD) - set(("style",)))
    OPTIONS_CUSTOM = ("bg", "fg")
    properties = OPTIONS_STANDARD + TTKButton.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    ro_properties = TTKButton.ro_properties + ("bg", "fg")
    class_ = atk.Button3d


_builder_uid = _plugin_uid + ".Button3d"
register_widget(
    _builder_uid, Button3dBO, "Button3d", ("ttk", _designer_tab_label), group=2
)
register_custom_property(
    _builder_uid, "bg", "colorentry", help=_("button color")
)

register_custom_property(_builder_uid, "fg", "colorentry", help=_("text color"))


class RadiobuttonBO(TTKRadiobutton):
    OPTIONS_STANDARD = tuple(
        set(TTKRadiobutton.OPTIONS_STANDARD) - set(("style",))
    )
    OPTIONS_CUSTOM = (
        "bg",
        "fg",
        "ind_bg",
        "ind_mark_color",
        "ind_outline_color",
        "font",
    )
    properties = (
        OPTIONS_STANDARD + TTKRadiobutton.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = (
        TTKRadiobutton.ro_properties + OPTIONS_CUSTOM + ("text", "value")
    )
    class_ = atk.Radiobutton


_builder_uid = _plugin_uid + ".Radiobutton"
register_widget(
    _builder_uid,
    RadiobuttonBO,
    "Radiobutton",
    ("ttk", _designer_tab_label),
    group=2,
)
register_custom_property(
    _builder_uid,
    "bg",
    "colorentry",
    help=_('background color "should match parent bg"'),
)
register_custom_property(_builder_uid, "fg", "colorentry", help=_("text color"))
register_custom_property(
    _builder_uid,
    "ind_bg",
    "colorentry",
    help=_('indicator ring background "fill color"'),
)
register_custom_property(
    _builder_uid,
    "ind_outline_color",
    "colorentry",
    help=_("indicator outline / ring color"),
)
register_custom_property(
    _builder_uid, "ind_mark_color", "colorentry", help=_("check mark color")
)
register_custom_property(
    _builder_uid,
    "font",
    "fontentry",
)


class CheckbuttonBO(TKCheckbutton):
    OPTIONS_CUSTOM = (
        "box_color",
        "check_mark_color",
        "text_color",
    )
    properties = (
        TKCheckbutton.OPTIONS_STANDARD
        + TKCheckbutton.OPTIONS_SPECIFIC
        + OPTIONS_CUSTOM
    )
    ro_properties = TKCheckbutton.ro_properties + OPTIONS_CUSTOM
    class_ = atk.Checkbutton


_builder_uid = _plugin_uid + ".Checkbutton"
register_widget(
    _builder_uid,
    CheckbuttonBO,
    "Checkbutton",
    ("ttk", _designer_tab_label),
    group=2,
)

register_custom_property(
    _builder_uid,
    "box_color",
    "colorentry",
)

register_custom_property(
    _builder_uid,
    "check_mark_color",
    "colorentry",
)

register_custom_property(
    _builder_uid,
    "text_color",
    "colorentry",
)



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk


import fixpath
import pygubu
import support


class TestOptionMenu(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_optionmenu.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("mainwindow")

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        optionmenu = self.builder.get_object("optionmenu1")
        self.assertIsInstance(optionmenu, tk.OptionMenu)
        self.widget.destroy()

    def test_no_variable_defined(self):
        optionmenu2 = self.builder.get_object("optionmenu2")
        self.assertIsInstance(optionmenu2, tk.OptionMenu)
        self.widget.destroy()



import tkinter as tk
from pygubu.i18n import _
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.ttk.ttkstdwidgets import TTKFrame
from ttkwidgets.frames import ScrolledFrame, ToggledFrame
from ..ttkwidgets import _designer_tab_label, _plugin_uid


class ScrolledFrameBO(TTKFrame):
    class_ = ScrolledFrame
    container = True
    container_layout = True
    OPTIONS_CUSTOM = (
        "compound",
        "canvaswidth",
        "canvasheight",
        "canvasborder",
        "autohidescrollbar",
    )
    properties = (
        TTKFrame.OPTIONS_STANDARD + TTKFrame.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = TTKFrame.ro_properties + OPTIONS_CUSTOM

    def get_child_master(self):
        return self.widget.interior

    def _process_property_value(self, pname, value):
        if pname == "autohidescrollbar":
            return tk.getboolean(value)
        if pname == "compound":
            return tk.LEFT if value == "left" else tk.RIGHT
        return super(ScrolledFrameBO, self)._process_property_value(
            pname, value
        )

    def code_child_master(self):
        return f"{self.code_identifier()}.interior"

    def _code_process_property_value(self, targetid, pname, value):
        if pname == "autohidescrollbar":
            return tk.getboolean(value)
        return super()._code_process_property_value(targetid, pname, value)


_builder_uid = f"{_plugin_uid}.ScrolledFrame"
register_widget(
    _builder_uid,
    ScrolledFrameBO,
    "ScrolledFrame",
    ("ttk", _designer_tab_label),
    group=0,
)

register_custom_property(
    _builder_uid,
    "compound",
    "choice",
    default_value=tk.RIGHT,
    values=("", tk.LEFT, tk.RIGHT),
    state="readonly",
    help=_("side the scrollbar should be on"),
)
register_custom_property(
    _builder_uid, "canvaswidth", "dimensionentry", default_value=400
)
register_custom_property(
    _builder_uid, "canvasheight", "dimensionentry", default_value=400
)
register_custom_property(_builder_uid, "canvasborder", "dimensionentry")
register_custom_property(
    _builder_uid,
    "autohidescrollbar",
    "choice",
    values=("", "true", "false"),
    state="readonly",
)


class ToggledFrameBO(TTKFrame):
    class_ = ToggledFrame
    container = True
    OPTIONS_CUSTOM = (
        "compound",
        "width",
        "text",
    )
    properties = (
        TTKFrame.OPTIONS_STANDARD + TTKFrame.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = TTKFrame.ro_properties + OPTIONS_CUSTOM

    def get_child_master(self):
        return self.widget.interior

    def _process_property_value(self, pname, value):
        final_value = value
        if pname in ("autohidescrollbar",):
            final_value = tk.getboolean(value)
        else:
            final_value = super(ToggledFrameBO, self)._process_property_value(
                pname, value
            )
        return final_value

    def code_child_master(self):
        return f"{self.code_identifier()}.interior"


_builder_uid = f"{_plugin_uid}.ToggledFrame"
register_widget(
    _builder_uid,
    ToggledFrameBO,
    "ToggledFrame",
    ("ttk", _designer_tab_label),
    group=0,
)

register_custom_property(
    _builder_uid,
    "compound",
    "choice",
    default_value=tk.RIGHT,
    help=_("position of the toggle arrow compared to the text"),
    values=("", tk.TOP, tk.BOTTOM, tk.LEFT, tk.RIGHT, tk.CENTER, tk.NONE),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "width",
    "naturalnumber",
    help=_("width of the closed ToggledFrame (in characters)"),
)
register_custom_property(
    _builder_uid,
    "text",
    "entry",
    help=_("text to display next to the toggle arrow"),
)



import tkinter as tk
import math
from ..photodraw import Draw
from .config import (
    PRIMARY,
    SECONDARY,
    SUCCESS,
    INFO,
    WARNING,
    DANGER,
    LIGHT,
    DARK,
    TTK_CLAM,
    TTK_ALT,
    TTK_DEFAULT,
)


class AssetCreator:
    def __init__(self, theme_builder):
        self.builder = theme_builder
        self.theme_images = {}

    @property
    def colors(self):
        return self.builder.colors

    @property
    def colorutil(self):
        return self.builder.colorutil

    @property
    def tk_master(self):
        return self.builder.tk_master

    @property
    def is_light_theme(self):
        return self.builder.is_light_theme

    def create_separator_assets(self, size, color):
        ssize = self.builder.scale_size(size)
        w = ssize[0]
        h = ssize[1]

        draw = Draw(self.tk_master)
        h_img = draw.canvas_new(width=w, height=h)
        draw.rectangle(0, 0, w, h, fill=color)
        v_img = draw.canvas_new(width=h, height=w)
        draw.rectangle(0, 0, h, w, fill=color)
        self.theme_images[h_img.name] = h_img
        self.theme_images[v_img.name] = v_img

        return (h_img.name, v_img.name)

    def create_sizegrip_assets(self, color):
        """Create image assets used to build the sizegrip style.

        Parameters:

            color (str):
                The color _value_ used to draw the image.

        Returns:

            str:
                The PhotoImage name.
        """
        box = self.builder.scale_size(1)
        pad = box * 2
        chunk = box + pad  # 4

        w = chunk * 3 + pad
        h = chunk * 3 + pad

        # size = [w, h]
        draw = Draw(self.tk_master)
        canvas = draw.canvas_new(width=w, height=h)
        draw.rectangle(chunk * 2 + pad, pad, chunk * 3, chunk, fill=color)
        draw.rectangle(
            chunk * 2 + pad, chunk + pad, chunk * 3, chunk * 2, fill=color
        )
        draw.rectangle(
            chunk * 2 + pad,
            chunk * 2 + pad,
            chunk * 3,
            chunk * 3,
            fill=color,
        )
        draw.rectangle(
            chunk + pad, chunk + pad, chunk * 2, chunk * 2, fill=color
        )
        draw.rectangle(
            chunk + pad, chunk * 2 + pad, chunk * 2, chunk * 3, fill=color
        )
        draw.rectangle(pad, chunk * 2 + pad, chunk, chunk * 3, fill=color)
        self.theme_images[canvas.name] = canvas
        return canvas.name

    def create_radiobutton_assets(self, colorname=None):
        """Create the image assets used to build the radiobutton style.

        Parameters:

            colorname (str):

        Returns:

            Tuple[str]:
                A tuple of PhotoImage names
        """
        prime_color = self.colors.get_color(colorname)
        on_fill = prime_color
        off_fill = self.colors.bg
        on_indicator = self.colors.selectfg
        size = self.builder.scale_size([14, 14])

        draw = Draw(self.tk_master)
        off_border = self.colorutil.make_transparent(
            0.4, self.colors.fg, self.colors.bg
        )
        disabled = self.colorutil.make_transparent(
            0.3, self.colors.fg, self.colors.bg
        )

        if self.is_light_theme:
            if colorname == LIGHT:
                on_indicator = self.colors.dark

        canvas_size = 134
        canvas = draw.canvas_new(width=canvas_size, height=canvas_size)

        # radio off
        x1 = y1 = 1
        x2 = y2 = 133
        stroke = 16
        draw.circle(
            x1, y1, x2, y2, fill=off_fill, outline=off_border, stroke=stroke
        )
        scale_factor = size[0] / canvas_size
        off_img = draw.canvas_scale(scale_factor)
        off_name = off_img.name
        self.theme_images[off_name] = off_img

        # radio on
        draw.canvas_blank()
        if colorname == LIGHT and self.is_light_theme:
            draw.circle(x1, y1, x2, y2, outline=off_border, stroke=stroke)
        else:
            draw.circle(x1, y1, x2, y2, fill=on_fill)
        ix1 = iy1 = 40
        ix2 = iy2 = 94
        draw.circle(ix1, iy1, ix2, iy2, fill=on_indicator)
        on_img = draw.canvas_scale(scale_factor)
        on_name = on_img.name
        self.theme_images[on_name] = on_img

        # radio on/disabled
        draw.canvas_blank()
        if colorname == LIGHT and self.is_light_theme:
            draw.circle(x1, y1, x2, y2, outline=off_border, stroke=stroke)
        else:
            draw.circle(x1, y1, x2, y2, fill=disabled)
        draw.circle(ix1, iy1, ix2, iy2, fill=off_fill)
        on_dis_img = draw.canvas_scale(scale_factor)
        on_disabled_name = on_dis_img.name
        self.theme_images[on_disabled_name] = on_dis_img

        # radio disabled
        draw.canvas_blank()
        stroke_3 = 10
        draw.circle(
            x1, y1, x2, y2, outline=disabled, stroke=stroke_3, fill=off_fill
        )
        disabled_img = draw.canvas_scale(scale_factor)
        disabled_name = disabled_img.name
        self.theme_images[disabled_name] = disabled_img
        del canvas

        return off_name, on_name, disabled_name, on_disabled_name

    def create_scrollbar_assets(self, thumbcolor, pressed, active):
        """Create the image assets used to build the standard scrollbar
        style.

        Parameters:

            thumbcolor (str):
                The primary color value used to color the thumb.

            pressed (str):
                The color value to use when the thumb is pressed.

            active (str):
                The color value to use when the thumb is active or
                hovered.
        """
        vsize = self.builder.scale_size([9, 28])
        hsize = self.builder.scale_size([28, 9])

        draw = Draw(self.tk_master)

        def create_image(size, color):
            # x = size[0] * 10
            # y = size[1] * 10
            img = draw.canvas_new()
            draw.rectangle(0, 0, size[0], size[1], fill=color)
            self.theme_images[img.name] = img
            return img

        # create images
        h_normal_img = create_image(hsize, thumbcolor)
        h_pressed_img = create_image(hsize, pressed)
        h_active_img = create_image(hsize, active)

        v_normal_img = create_image(vsize, thumbcolor)
        v_pressed_img = create_image(vsize, pressed)
        v_active_img = create_image(vsize, active)

        return (
            h_normal_img,
            h_pressed_img,
            h_active_img,
            v_normal_img,
            v_pressed_img,
            v_active_img,
        )

    def create_scale_assets(self, colorname=None, size=14):
        """Create the assets used for the ttk.Scale widget.

        The slider handle is automatically adjusted to fit the
        screen resolution.

        Parameters:

            colorname (str):
                The color label.

            size (int):
                The size diameter of the slider circle; default=16.

        Returns:

            Tuple[str]:
                A tuple of PhotoImage names to be used in the image
                layout when building the style.
        """
        size = self.builder.scale_size(size)
        draw = Draw(self.tk_master)

        if self.is_light_theme:
            disabled_color = self.colors.border
            if colorname == LIGHT:
                track_color = self.colors.bg
            else:
                track_color = self.colors.light
        else:
            disabled_color = self.colors.selectbg
            track_color = self.colorutil.update_hsv(
                self.colors.selectbg, vd=-0.2
            )

        if any([colorname is None, not colorname]):
            normal_color = self.colors.primary
        else:
            normal_color = self.colors.get_color(colorname)

        pressed_color = self.colorutil.update_hsv(normal_color, vd=-0.1)
        hover_color = self.colorutil.update_hsv(normal_color, vd=0.1)

        canvas_size = 100
        scale_factor = size / canvas_size

        # normal state
        canvas = draw.canvas_new(width=canvas_size, height=canvas_size)
        x1 = y1 = 0
        x2 = y2 = 95
        draw.circle(x1, y1, x2, y2, fill=normal_color)
        normal_img = draw.canvas_scale(scale_factor)
        normal_name = normal_img.name
        self.theme_images[normal_name] = normal_img

        # pressed state
        draw.canvas_blank()
        draw.circle(x1, y1, x2, y2, fill=pressed_color)
        pressed_img = draw.canvas_scale(scale_factor)
        pressed_name = pressed_img.name
        self.theme_images[pressed_name] = pressed_img

        # hover state
        draw.canvas_blank()
        draw.circle(x1, y1, x2, y2, fill=hover_color)
        hover_img = draw.canvas_scale(scale_factor)
        hover_name = hover_img.name
        self.theme_images[hover_name] = hover_img

        # disabled state
        draw.canvas_blank()
        draw.circle(x1, y1, x2, y2, fill=disabled_color)
        disabled_img = draw.canvas_scale(scale_factor)
        disabled_name = disabled_img.name
        self.theme_images[disabled_name] = disabled_img
        del canvas

        def create_image(size, color):
            img = draw.canvas_new()
            draw.rectangle(0, 0, size[0], size[1], fill=color)
            self.theme_images[img.name] = img
            return img

        # vertical track ??
        h_track_img = create_image(
            self.builder.scale_size((40, 5)), track_color
        )
        h_track_name = h_track_img.name

        # horizontal track ??
        v_track_img = create_image(
            self.builder.scale_size((5, 40)), track_color
        )
        v_track_name = v_track_img.name

        return (
            normal_name,
            pressed_name,
            hover_name,
            disabled_name,
            h_track_name,
            v_track_name,
        )

    def create_checkbutton_assets(self, colorname=None):
        """Create the image assets used to build the standard
        checkbutton style.

        Parameters:

            colorname (str):
                The color label used to style the widget.

        Returns:

            Tuple[str]:
                A tuple of PhotoImage names.
        """

        draw = Draw(self.tk_master)
        prime_color = self.colors.get_color(colorname)
        on_border = prime_color
        on_fill = prime_color
        off_fill = self.colors.bg
        off_border = self.colors.selectbg
        off_border = self.colorutil.make_transparent(
            0.4, self.colors.fg, self.colors.bg
        )
        disabled_fg = self.colorutil.make_transparent(
            0.3, self.colors.fg, self.colors.bg
        )

        if colorname == LIGHT:
            check_color = self.colors.dark
            on_border = check_color
        elif colorname == DARK:
            check_color = self.colors.light
            on_border = check_color
        else:
            check_color = self.colors.selectfg

        size = self.builder.scale_size([14, 14])

        canvas_size = 134
        scale_factor = size[0] / canvas_size

        # checkbutton off
        canvas = draw.canvas_new(width=canvas_size, height=canvas_size)
        x1 = y1 = 2
        x2 = y2 = 132
        rec_radio = 32
        stroke = 18

        draw.rounded_rectangle(
            x1,
            y1,
            x2,
            y2,
            radio=rec_radio,
            fill=off_fill,
            outline=off_border,
            stroke=stroke,
        )
        off_img = draw.canvas_scale(scale_factor)
        off_name = off_img.name
        self.theme_images[off_name] = off_img

        # checkbutton on
        draw.canvas_blank()
        rec_stroke_3 = 10
        draw.rounded_rectangle(
            x1,
            y1,
            x2,
            y2,
            radio=rec_radio,
            stroke=rec_stroke_3,
            outline=on_border,
            fill=on_fill,
        )

        shape = [
            [(45, 85), (35, 85), (35, 100), (45, 100)],
            [(48, 87), (66, 60), (82, 60), (63, 87)],
            [(106, 27), (96, 27), (89, 30), (106, 30)],
            [(42, 75), (34, 75), (28, 80), (44, 80)],
            [(44, 80), (28, 80), (30, 90), (49, 90)],
            [(48, 90), (30, 90), (37, 107), (48, 107)],
            [(63, 87), (46, 87), (46, 107), (50, 107)],
            [(66, 60), (89, 30), (106, 30), (82, 60)],
        ]

        for poly in shape:
            draw.convex_poly(poly, check_color)
        # for tri in shape:
        #    draw.triangle_filled(*tri, check_color)

        on_img = draw.canvas_scale(scale_factor)
        on_name = on_img.name
        self.theme_images[on_name] = on_img

        # checkbutton on/disabled
        draw.canvas_blank()
        draw.rounded_rectangle(
            x1,
            y1,
            x2,
            y2,
            radio=rec_radio,
            stroke=rec_stroke_3,
            outline=disabled_fg,
            fill=disabled_fg,
        )

        for poly in shape:
            draw.convex_poly(poly, check_color)
        # for tri in shape:
        #     draw.triangle_filled(*tri, off_fill)

        on_dis_img = draw.canvas_scale(scale_factor)
        on_dis_name = on_dis_img.name
        self.theme_images[on_dis_name] = on_dis_img

        # checkbutton alt
        draw.canvas_blank()
        draw.rounded_rectangle(
            x1,
            y1,
            x2,
            y2,
            radio=rec_radio,
            stroke=rec_stroke_3,
            outline=on_border,
            fill=on_fill,
        )
        line_coords = [36, 58, 100, 70]
        draw.rectangle(*line_coords, fill=check_color)

        alt_img = draw.canvas_scale(scale_factor)
        alt_name = alt_img.name
        self.theme_images[alt_name] = alt_img

        # checkbutton alt/disabled
        draw.canvas_blank()
        draw.rounded_rectangle(
            x1,
            y1,
            x2,
            y2,
            radio=rec_radio,
            stroke=rec_stroke_3,
            outline=disabled_fg,
            fill=disabled_fg,
        )
        draw.rectangle(*line_coords, fill=off_fill)
        alt_dis_img = draw.canvas_scale(scale_factor)
        alt_dis_name = alt_dis_img.name
        self.theme_images[alt_dis_name] = alt_dis_img

        # checkbutton disabled
        draw.canvas_blank()
        draw.rounded_rectangle(
            x1,
            y1,
            x2,
            y2,
            radio=rec_radio,
            stroke=rec_stroke_3,
            outline=disabled_fg,
        )
        disabled_img = draw.canvas_scale(scale_factor)
        disabled_name = disabled_img.name
        self.theme_images[disabled_name] = disabled_img
        del canvas

        return (
            off_name,
            on_name,
            disabled_name,
            alt_name,
            on_dis_name,
            alt_dis_name,
        )

    def create_striped_progressbar_assets(self, thickness, colorname=None):
        """Create the striped progressbar image and return as a
        `PhotoImage`

        Parameters:

            colorname (str):
                The color label used to style the widget.

        Returns:

            Tuple[str]:
                A list of photoimage names.
        """
        if any([colorname is None, colorname == ""]):
            barcolor = self.colors.primary
        else:
            barcolor = self.colors.get_color(colorname)

        draw = Draw(self.tk_master)
        # calculate value of the light color
        brightness = self.colorutil.rgb_to_hsv(
            *self.colorutil.hex_to_rgb(barcolor)
        )[2]
        if brightness < 0.4:
            value_delta = 0.3
        elif brightness > 0.8:
            value_delta = 0
        else:
            value_delta = 0.1

        barcolor_light = self.colorutil.update_hsv(
            barcolor, sd=-0.2, vd=value_delta
        )

        # horizontal progressbar
        canvas_size = 100
        canvas = draw.canvas_new(width=canvas_size, height=canvas_size)
        draw.rectangle(0, 0, 100, 100, fill=barcolor_light)
        shape = [(0, 0), (48, 0), (100, 52), (100, 100)]
        draw.convex_poly(shape, barcolor)
        shape = [(0, 52), (48, 100), (0, 100)]
        draw.convex_poly(shape, barcolor)

        scale_factor = thickness / canvas_size
        h_img = draw.canvas_scale(scale_factor)
        h_name = h_img.name

        v_img = h_img.copy()
        v_name = v_img.name
        del canvas
        draw.canvas = v_img
        draw.canvas_rotate(90)

        self.theme_images[h_name] = h_img
        self.theme_images[v_name] = v_img
        return h_name, v_name

    def create_square_toggle_assets(self, colorname=None):
        """Create the image assets used to build a square toggle
        style.

        Parameters:

            colorname (str):
                The color label used to style the widget.

        Returns:

            Tuple[str]:
                A tuple of PhotoImage names.
        """
        size = self.builder.scale_size([24, 15])
        if any([colorname is None, colorname == ""]):
            colorname = PRIMARY

        # set default style color values
        prime_color = self.colors.get_color(colorname)
        on_border = prime_color
        on_indicator = self.colors.selectfg
        on_fill = prime_color
        off_fill = self.colors.bg

        draw = Draw(self.tk_master)

        disabled_fg = self.colorutil.make_transparent(
            0.3, self.colors.fg, self.colors.bg
        )
        off_border = self.colorutil.make_transparent(
            0.4, self.colors.fg, self.colors.bg
        )
        off_indicator = self.colorutil.make_transparent(
            0.4, self.colors.fg, self.colors.bg
        )

        # override defaults for light and dark colors
        if colorname == LIGHT:
            on_border = self.colors.dark
            on_indicator = on_border
        elif colorname == DARK:
            on_border = self.colors.light
            on_indicator = on_border

        canvasw = 226
        canvash = 130
        sfw = size[0] / canvasw
        sfh = size[1] / canvash
        stroke6 = 9

        # toggle off
        canvas = draw.canvas_new(width=canvasw, height=canvash)
        draw.rectangle(
            1, 1, 225, 129, outline=off_border, stroke=stroke6, fill=off_fill
        )
        draw.rectangle(20, 20, 110, 110, fill=off_indicator)

        off_img = draw.canvas_scale(sfw, sfh)
        off_name = off_img.name
        self.theme_images[off_name] = off_img

        # toggle on
        draw.canvas_blank()
        draw.rectangle(
            1, 1, 225, 129, outline=on_border, stroke=stroke6, fill=on_fill
        )
        draw.rectangle(20, 20, 110, 110, fill=on_indicator)
        draw.canvas_rotate(180)
        on_img = draw.canvas_scale(sfw, sfh)
        on_name = on_img.name
        self.theme_images[on_name] = on_img

        # toggle disabled
        draw.canvas_blank()
        draw.rectangle(1, 1, 225, 129, outline=disabled_fg, stroke=stroke6)
        draw.rectangle(20, 20, 110, 110, fill=disabled_fg)
        disabled_img = draw.canvas_scale(sfw, sfh)
        disabled_name = disabled_img.name
        self.theme_images[disabled_name] = disabled_img

        # toggle on / disabled
        draw.canvas_blank()
        draw.rectangle(
            1, 1, 225, 129, outline=disabled_fg, stroke=stroke6, fill=off_fill
        )
        draw.rectangle(20, 20, 110, 110, fill=disabled_fg)
        draw.canvas_rotate(180)
        on_dis_img = draw.canvas_scale(sfw, sfh)
        on_disabled_name = on_dis_img.name
        self.theme_images[on_disabled_name] = on_dis_img

        del canvas
        return off_name, on_name, disabled_name, on_disabled_name

    def create_round_toggle_assets(self, colorname=None):
        """Create image assets for the round toggle style.

        Parameters:

            colorname (str):
                The color label assigned to the colors property.

        Returns:

            Tuple[str]:
                A tuple of PhotoImage names.
        """
        size = self.builder.scale_size([24, 15])

        if any([colorname is None, colorname == ""]):
            colorname = PRIMARY

        # set default style color values
        prime_color = self.colors.get_color(colorname)
        on_border = prime_color
        on_indicator = self.colors.selectfg
        on_fill = prime_color
        off_fill = self.colors.bg

        draw = Draw(self.tk_master)

        disabled_fg = self.colorutil.make_transparent(
            0.3, self.colors.fg, self.colors.bg
        )
        off_border = self.colorutil.make_transparent(
            0.4, self.colors.fg, self.colors.bg
        )
        off_indicator = self.colorutil.make_transparent(
            0.4, self.colors.fg, self.colors.bg
        )

        # override defaults for light and dark colors
        if colorname == LIGHT:
            on_border = self.colors.dark
            on_indicator = on_border
        elif colorname == DARK:
            on_border = self.colors.light
            on_indicator = on_border

        # specific
        canvasw = 226
        canvash = 130
        sfw = size[0] / canvasw
        sfh = size[1] / canvash
        stroke6 = 9
        radius = int(128 / 2)

        # toggle off
        rec_coords = (1, 1, 225, 129)
        ellip_coords = (20, 18, 112, 110)
        canvas = draw.canvas_new(width=canvasw, height=canvash)
        draw.rounded_rectangle(
            *rec_coords,
            radio=radius,
            outline=off_border,
            stroke=stroke6,
            fill=off_fill,
        )
        draw.circle(*ellip_coords, fill=off_indicator)

        off_img = draw.canvas_scale(sfw, sfh)
        off_name = off_img.name
        self.theme_images[off_name] = off_img

        # toggle on
        draw.canvas_blank()
        draw.rounded_rectangle(
            *rec_coords,
            radio=radius,
            outline=on_border,
            stroke=stroke6,
            fill=on_fill,
        )
        draw.circle(*ellip_coords, fill=on_indicator)
        draw.canvas_rotate(180)

        on_img = draw.canvas_scale(sfw, sfh)
        on_name = on_img.name
        self.theme_images[on_name] = on_img

        # toggle on / disabled
        draw.canvas_blank()
        draw.rounded_rectangle(
            *rec_coords,
            radio=radius,
            outline=disabled_fg,
            stroke=stroke6,
            fill=off_fill,
        )
        draw.circle(*ellip_coords, fill=disabled_fg)
        draw.canvas_rotate(180)

        on_dis_img = draw.canvas_scale(sfw, sfh)
        on_disabled_name = on_dis_img.name
        self.theme_images[on_disabled_name] = on_dis_img

        # toggle disabled
        draw.canvas_blank()
        draw.rounded_rectangle(
            *rec_coords, radio=radius, outline=disabled_fg, stroke=stroke6
        )
        draw.circle(*ellip_coords, fill=disabled_fg)

        disabled_img = draw.canvas_scale(sfw, sfh)
        disabled_name = disabled_img.name
        self.theme_images[disabled_name] = disabled_img
        del canvas
        return off_name, on_name, disabled_name, on_disabled_name

    def create_round_scrollbar_assets(self, thumbcolor, pressed, active):
        """Create image assets to be used when building the round
        scrollbar style.

        Parameters:

            thumbcolor (str):
                The color value of the thumb in normal state.

            pressed (str):
                The color value to use when the thumb is pressed.

            active (str):
                The color value to use when the thumb is active or
                hovered.
        """
        vsize = self.builder.scale_size([9, 28])
        hsize = self.builder.scale_size([28, 9])

        draw_v = Draw(self.tk_master)
        vw = vsize[0] * 10
        vh = vsize[1] * 10
        vsfw = vsize[0] / vw
        vsfh = vsize[1] / vh
        vradius = min([vw, vh]) // 2
        draw_v.canvas_new(width=vw, height=vh)

        hw = hsize[0]
        hh = hsize[1]
        hsfw = hsize[0] / hw
        hsfh = hsize[1] / hh
        hradius = min([hw, hh]) // 2
        draw_h = Draw(self.tk_master)
        draw_h.canvas_new(width=hsize[0], height=hsize[1])

        def vrounded_rect(color):
            draw_v.canvas_blank()
            draw_v.rounded_rectangle(
                0, 0, vw - 1, vh - 1, radio=vradius, fill=color
            )
            img = draw_v.canvas_scale(vsfw, vsfh)
            self.theme_images[img.name] = img
            return img.name

        def hrounded_rect(color):
            draw_h.canvas_blank()
            draw_h.rounded_rectangle(
                0, 0, hw - 1, hh - 1, radio=hradius, fill=color
            )
            img = draw_h.canvas_scale(hsfw, hsfh)
            self.theme_images[img.name] = img
            return img.name

        # create images
        h_normal_img = hrounded_rect(thumbcolor)
        h_pressed_img = hrounded_rect(pressed)
        h_active_img = hrounded_rect(active)

        v_normal_img = vrounded_rect(thumbcolor)
        v_pressed_img = vrounded_rect(pressed)
        v_active_img = vrounded_rect(active)

        del draw_h.canvas
        del draw_v.canvas

        return (
            h_normal_img,
            h_pressed_img,
            h_active_img,
            v_normal_img,
            v_pressed_img,
            v_active_img,
        )



import tkinter as tk
import pygubu.forms.ttkwidget as ttkwidget
import pygubu.plugins.tk.tkstdwidgets as tkw
import pygubu.plugins.ttk.ttkstdwidgets as ttkw

from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from pygubu.i18n import _
from .base import (
    WidgetBOMixin,
    _plugin_uid as base_plugin_uid,
    _designer_tabname,
)
from pygubu.plugins.pygubu.scrollbarhelper import TTKSBHelperBO
from .tkwidgetbo import _tk_text_builder_uid

# Groups for ordering buttons in designer palette.
GROUP0: int = 0
GROUP1: int = 10
GROUP2: int = 20
GROUP3: int = 30

_plugin_uid = f"{base_plugin_uid}.ttkwidget"
_designer_tabs = ("ttk", _designer_tabname)

# Register text as child of TTKSBHelperBO
TTKSBHelperBO.add_allowed_child(_tk_text_builder_uid)


class FrameFormBuilderBO(WidgetBOMixin, ttkw.TTKFrame):
    class_ = ttkwidget.FrameFormBuilder
    properties = ttkw.TTKFrame.properties + WidgetBOMixin.base_properties
    ro_properties = ttkw.TTKFrame.ro_properties + WidgetBOMixin.base_properties


_builder_uid = f"{_plugin_uid}.FrameFormBuilder"
register_widget(
    _builder_uid,
    FrameFormBuilderBO,
    "FrameFormBuilder",
    _designer_tabs,
    group=GROUP0,
)


class EntryBO(WidgetBOMixin, ttkw.TTKEntry):
    class_ = ttkwidget.Entry
    properties = ttkw.TTKEntry.properties + WidgetBOMixin.base_properties
    ro_properties = ttkw.TTKEntry.ro_properties + WidgetBOMixin.base_properties


_builder_uid = f"{_plugin_uid}.Entry"
register_widget(
    _builder_uid,
    EntryBO,
    "Entry",
    _designer_tabs,
    group=GROUP0,
)


class LabelWidgetInfoBO(WidgetBOMixin, ttkw.TTKLabel):
    class_ = ttkwidget.LabelWidgetInfo
    properties = ttkw.TTKLabel.properties + WidgetBOMixin.base_properties
    ro_properties = ttkw.TTKLabel.ro_properties + WidgetBOMixin.base_properties


_builder_uid = f"{_plugin_uid}.LabelWidgetInfo"
register_widget(
    _builder_uid,
    LabelWidgetInfoBO,
    "LabelWidgetInfo",
    _designer_tabs,
    group=GROUP1,
)
register_custom_property(_builder_uid, "field_name", "fieldname_selector")


class LabelBO(WidgetBOMixin, ttkw.TTKLabel):
    class_ = ttkwidget.Label
    properties = ttkw.TTKLabel.properties + WidgetBOMixin.base_properties
    ro_properties = ttkw.TTKLabel.ro_properties + WidgetBOMixin.base_properties


_builder_uid = f"{_plugin_uid}.Label"
register_widget(
    _builder_uid,
    LabelBO,
    "Label",
    _designer_tabs,
    group=GROUP0,
)


class CheckbuttonBO(WidgetBOMixin, ttkw.TTKCheckbutton):
    class_ = ttkwidget.Checkbutton
    properties = ttkw.TTKCheckbutton.properties + WidgetBOMixin.base_properties
    ro_properties = (
        ttkw.TTKCheckbutton.ro_properties + WidgetBOMixin.base_properties
    )


_builder_uid = f"{_plugin_uid}.Checkbutton"
register_widget(
    _builder_uid,
    CheckbuttonBO,
    "Checkbutton",
    _designer_tabs,
    group=GROUP0,
)


class ComboboxBO(WidgetBOMixin, ttkw.TTKCombobox):
    class_ = ttkwidget.Combobox
    properties = ttkw.TTKCombobox.properties + WidgetBOMixin.base_properties
    ro_properties = (
        ttkw.TTKCombobox.ro_properties + WidgetBOMixin.base_properties
    )


_builder_uid = f"{_plugin_uid}.Combobox"
register_widget(
    _builder_uid,
    ComboboxBO,
    "Combobox",
    _designer_tabs,
    group=GROUP0,
)

register_custom_property(
    _builder_uid,
    "state",
    "choice",
    values=("", "normal", "readonly", "disabled"),
    state="readonly",
)



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk

import fixpath
import pygubu
import support


class TestButtonCommand(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_button_command.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("mainwindow")
        self.button1 = builder.get_object("button1")
        self.button2 = builder.get_object("button2")

    def tearDown(self):
        support.root_withdraw()

    def test_command_simple(self):
        success = []

        class AnObject:
            def button1_clicked(self):
                success.append(1)

            def on_button_clicked(self):
                pass

        cbobj = AnObject()
        self.builder.connect_callbacks(cbobj)
        self.button1.invoke()
        self.widget.update()

        self.assertTrue(success)
        self.widget.destroy()

    def test_command_with_widgetid(self):
        success = []
        received_id = []

        class AnObject:
            def button1_clicked(self):
                pass

            def on_button_clicked(self, widgetid):
                success.append(1)
                received_id.append(widgetid)

        cbobj = AnObject()
        self.builder.connect_callbacks(cbobj)
        self.button2.invoke()
        self.widget.update()

        self.assertTrue(success)
        self.assertEqual(received_id[0], "button2")
        self.widget.destroy()

    # def test_command_generate_event(self):
    # success = []

    # class AnObject:
    # def button1_clicked(self):
    # pass
    # def on_button_clicked(self):
    # pass
    # def catch_button_event(self, event):
    # success.append(1)

    # cbobj = AnObject()
    # self.button3.bind('<<MyButtonEvent>>', cbobj.catch_button_event)
    # self.builder.connect_callbacks(cbobj)
    # self.button3.invoke()
    # self.widget.update()

    # self.assertTrue(success)
    # self.widget.destroy()



import tkinter as tk

from pygubu.api.v1 import BuilderObject, register_widget
from .basehelpers import (
    ToplevelPreviewFactory,
    ToplevelPreviewMixin,
    ToplevelPreviewBaseBO,
)


ToplevelFramePreview = ToplevelPreviewFactory(
    "ToplevelFramePreview",
    (ToplevelPreviewMixin, tk.Frame, object),
    {},
)


class ToplevelFramePreviewBO(ToplevelPreviewBaseBO):
    class_ = ToplevelFramePreview



# encoding: utf-8
import tkinter as tk
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from pygubu.plugins.ttk.ttkstdwidgets import TTKFrame
from pygubu.i18n import _
from pygubu.widgets.accordionframe import AccordionFrame


class AccordionFrameBO(TTKFrame):
    class_ = AccordionFrame
    allowed_children = ("pygubu.widgets.AccordionFrameGroup",)
    _img_properties = ("img_expand", "img_collapse")
    properties = TTKFrame.properties + _img_properties
    tkimage_properties = TTKFrame.tkimage_properties + _img_properties


_builder_uid = "pygubu.widgets.AccordionFrame"

register_widget(
    _builder_uid,
    AccordionFrameBO,
    "AccordionFrame",
    (_("Pygubu Widgets"), "ttk"),
)

for prop in AccordionFrameBO._img_properties:
    register_custom_property(_builder_uid, prop, "imageentry")


class AccordionFrameGroupBO(BuilderObject):
    allowed_parents = ("pygubu.widgets.AccordionFrame",)
    properties = ("label", "expanded", "compound", "style")
    layout_required = False
    container = True
    container_layout = True
    allow_bindings = False

    def realize(self, parent, extra_init_args: dict = None):
        args = self._get_init_args(extra_init_args)
        master = parent.get_child_master()
        self._accordion = parent.widget
        gid = self.wmeta.identifier
        self.widget = master.add_group(gid, **args)
        return self.widget

    def _process_property_value(self, pname, value):
        if pname == "expanded":
            return tk.getboolean(value)
        return super()._process_property_value(pname, value)

    def _set_property(self, target_widget, pname, value):
        if pname in self.properties:
            propvalue = self._process_property_value(pname, value)
            self._accordion.group_config(
                self.wmeta.identifier, **{pname: propvalue}
            )
        else:
            super()._set_property(target_widget, pname, value)


_builder_uid = "pygubu.widgets.AccordionFrameGroup"

register_widget(
    _builder_uid,
    AccordionFrameGroupBO,
    "AccordionFrame.Group",
    (_("Pygubu Widgets"), "ttk"),
)

register_custom_property(_builder_uid, "label", "entry")
register_custom_property(
    _builder_uid, "style", "ttkstylechoice", default_value="Toolbutton"
)
register_custom_property(
    _builder_uid,
    "expanded",
    "choice",
    values=("", "false", "true"),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "compound",
    "choice",
    values=(
        "",
        tk.LEFT,
        tk.RIGHT,
        tk.NONE,
    ),
    state="readonly",
)



import tkinter as tk
import pygubu.plugins.pygubu.combobox as cbb
import pygubu.forms.pygubuwidget as pygubuwidget

from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
    copy_custom_property,
)
from pygubu.i18n import _
from .base import (
    WidgetBOMixin,
    _plugin_uid as base_plugin_uid,
    _designer_tabname,
)


_plugin_uid = f"{base_plugin_uid}.pygubuwidget"
_designer_tabs = ("tk", _designer_tabname)


class PygubuComboboxBO(WidgetBOMixin, cbb.ComboboxBO):
    class_ = pygubuwidget.PygubuCombobox
    properties = cbb.ComboboxBuilder.properties + WidgetBOMixin.base_properties
    ro_properties = (
        cbb.ComboboxBuilder.ro_properties + WidgetBOMixin.base_properties
    )


_builder_uid = f"{_plugin_uid}.PygubuCombobox"
register_widget(
    _builder_uid,
    PygubuComboboxBO,
    "PygubuCombobox",
    _designer_tabs,
)

copy_custom_property(cbb._builder_id, "values", _builder_uid)
copy_custom_property(cbb._builder_id, "keyvariable", _builder_uid)
copy_custom_property(cbb._builder_id, "state", _builder_uid)



"""No UI forms"""

from .forms import FormBase
from .widgets import FieldWidget
from .fields import (
    CharField as CharFieldBase,
    IntegerField as IntegerFieldBase,
    ChoiceField as ChoiceFieldBase,
)


class Form(FormBase):
    pass


class NoUIWidget(FieldWidget):
    def __init__(self, *args, **kw):
        self._value = None
        super().__init__(*args, **kw)

    def wset_value(self, value):
        self._value = value

    def wget_value(self):
        return self._value

    def wmark_invalid(self, state: bool):
        print(f"Field {self._field.fname} invalid: {state}")

    def wis_disabled(self) -> bool:
        return False


class CharField(CharFieldBase, NoUIWidget):
    pass


class IntegerField(IntegerFieldBase, NoUIWidget):
    pass


# class NoUIChoiceWidget(NoUIWidget, ChoiceWidget):
#    pass


# class ChoiceField(ChoiceFieldBase, NoUIChoiceWidget):
#    pass



"""
Documentation, License etc.

@package pygubu.plugins.awesometkinter
"""
import tkinter as tk
from pygubu.i18n import _
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.tk.tkstdwidgets import TKFrame
from pygubu.plugins.ttk.ttkstdwidgets import TTKFrame
import awesometkinter as atk

from ..awesometkinter import _designer_tab_label, _plugin_uid


class Frame3dBO(TTKFrame):
    OPTIONS_STANDARD = tuple(set(TTKFrame.OPTIONS_STANDARD) - set(("style",)))
    OPTIONS_CUSTOM = ("bg",)
    properties = OPTIONS_STANDARD + TTKFrame.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    ro_properties = TTKFrame.ro_properties + ("bg",)
    class_ = atk.Frame3d


_builder_uid = _plugin_uid + ".Frame3d"
register_widget(
    _builder_uid, Frame3dBO, "Frame3d", ("ttk", _designer_tab_label), group=0
)

register_custom_property(
    _builder_uid, "bg", "colorentry", help=_("color of frame")
)


class ScrollableFrameBO(TKFrame):
    OPTIONS_STANDARD = tuple()
    OPTIONS_SPECIFIC = tuple()
    OPTIONS_CUSTOM = (
        "vscroll",
        "hscroll",
        "autoscroll",
        "bg",
        "sbar_fg",
        "sbar_bg",
        "vbar_width",
        "hbar_width",
    )
    ro_properties = OPTIONS_CUSTOM
    class_ = atk.ScrollableFrame

    def _process_property_value(self, pname, value):
        if pname in ("vscroll", "hscroll", "autoscroll"):
            return tk.getboolean(value)
        return super()._process_property_value(pname, value)

    def _code_process_property_value(self, targetid, pname, value):
        if pname in ("vscroll", "hscroll", "autoscroll"):
            return tk.getboolean(value)
        return super()._code_process_property_value(targetid, pname, value)


_builder_uid = _plugin_uid + ".ScrollableFrame"
register_widget(
    _builder_uid,
    ScrollableFrameBO,
    "ScrollableFrame",
    ("ttk", _designer_tab_label),
    group=0,
)

register_custom_property(
    _builder_uid,
    "vscroll",
    "choice",
    values=("", "false", "true"),
    state="readonly",
    help=_("use vertical scrollbar"),
)

register_custom_property(
    _builder_uid,
    "hscroll",
    "choice",
    values=("", "false", "true"),
    state="readonly",
    help=_("use horizontal scrollbar"),
)

register_custom_property(
    _builder_uid,
    "autoscroll",
    "choice",
    values=("", "false", "true"),
    state="readonly",
    help=_("auto scroll to bottom if new items added to frame"),
)

register_custom_property(
    _builder_uid, "bg", "colorentry", help=_("background color")
)

register_custom_property(
    _builder_uid, "sbar_fg", "colorentry", help=_("color of scrollbars' slider")
)

register_custom_property(
    _builder_uid,
    "sbar_bg",
    "colorentry",
    help=_("color of scrollbars' trough, default to frame's background"),
)

register_custom_property(
    _builder_uid,
    "vbar_width",
    "dimensionentry",
    help=_("vertical scrollbar width"),
)

register_custom_property(
    _builder_uid,
    "hbar_width",
    "dimensionentry",
    help=_("vertical scrollbar width"),
)



#!/usr/bin/python3
# Note: Template in development.
import tkinter as tk
import tkinter.ttk as ttk


class SlotIndicator(tk.Frame):
    def __init__(self, *args, side="n", **kw):
        super().__init__(*args, **kw)
        self.side = side


#
# Base class definition
#
class SlotFrameBase(ttk.Frame):
    def __init__(self, master=None, **kw):
        super().__init__(master, **kw)
        self.fn = SlotIndicator(
            self,
            cursor="top_side",
            background="#003ed9",
            height=15,
            width=15,
            side="n",
            name="fn",
        )
        self.fn.grid(column=0, columnspan=3, row=0, sticky="ew")
        self.fs = SlotIndicator(
            self,
            cursor="bottom_side",
            background="#003ed9",
            height=15,
            width=10,
            side="s",
            name="fs",
        )
        self.fs.grid(column=0, columnspan=3, row=2, sticky="ew")
        self.fe = SlotIndicator(
            self,
            cursor="right_side",
            background="#003ed9",
            height=15,
            width=15,
            side="e",
            name="fe",
        )
        self.fe.grid(column=2, row=1, sticky="ns")
        self.fw = SlotIndicator(
            self,
            cursor="left_side",
            background="#003ed9",
            height=15,
            width=15,
            side="w",
            name="fw",
        )
        self.fw.grid(column=0, row=1, sticky="ns")
        self.fcenter = ttk.Frame(self, name="fcenter")
        self.fcenter.configure(height=20, width=20)
        self.fcenter.grid(column=1, row=1, sticky="nsew")
        self.configure(height=200, width=200)
        # self.pack(side="top")
        # self.grid_propagate(0)
        self.rowconfigure(1, weight=1)
        self.columnconfigure(1, weight=1)


#
# Manual user code
#


class SlotFrame(SlotFrameBase):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        for i in self.iter_indicators():
            i.grid_remove()
        self._indicators_visible = False

    def iter_indicators(self) -> tk.Widget:
        for fi in (self.fn, self.fs, self.fw, self.fe):
            yield fi

    def indicators_visible(self, visible: bool):
        if self._indicators_visible != visible:
            self._indicators_visible = visible
            if visible:
                self.configure(padding=4)
            else:
                self.configure(padding=0)

            for f in self.iter_indicators():
                if visible:
                    f.grid()
                else:
                    f.grid_remove()



from pygubu.i18n import _
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.ttk.ttkstdwidgets import TTKLabel
from ttkwidgets import LinkLabel

from ..ttkwidgets import _designer_tab_label, _plugin_uid


class LinkLabelBO(TTKLabel):
    class_ = LinkLabel
    OPTIONS_CUSTOM = ("link", "normal_color", "hover_color", "clicked_color")
    properties = (
        TTKLabel.OPTIONS_STANDARD + TTKLabel.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )


_builder_uid = f"{_plugin_uid}.LinkLabel"
register_widget(
    _builder_uid,
    LinkLabelBO,
    "LinkLabel",
    ("ttk", _designer_tab_label),
    group=1,
)

register_custom_property(
    _builder_uid, "link", "entry", help=_("link to be opened")
)
register_custom_property(
    _builder_uid,
    "normal_color",
    "colorentry",
    help=_("text color when widget is created"),
)
register_custom_property(
    _builder_uid,
    "hover_color",
    "colorentry",
    help=_("text color when hovering over the widget"),
)
register_custom_property(
    _builder_uid,
    "clicked_color",
    "colorentry",
    help=_("text color when link is clicked"),
)



import os
import tkinter as tk
import customtkinter
from pathlib import Path
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from pygubu.i18n import _
from pygubu.plugins.tk.tkstdwidgets import TKFrame as TKFrameBO
from pygubu.utils.font import tkfontstr_to_dict
from pygubu.stockimage import StockImage
from ..customtkinter import _designer_tab_label, _plugin_uid
from customtkinter.windows.widgets.core_widget_classes import CTkBaseClass
from customtkinter import CTkFont, CTkImage
from PIL import Image, ImageTk


# Groups for ordering buttons in designer palette.
GCONTAINER = 0
GDISPLAY = 1
GINPUT = 2


_use_fixed_image_class = False

if os.getenv("PYGUBU_DESIGNER_RUNNING"):
    _use_fixed_image_class = True

    class CTKImageFix(CTkImage):
        "Fix loader for pygubu designer toplevel preview"

        def __init__(
            self,
            light_image: "Image.Image" = None,
            dark_image: "Image.Image" = None,
            size=(20, 20),
            tk_master=None,
        ):
            super().__init__(light_image, dark_image, size)
            self._tk_master = tk_master

        def _get_scaled_light_photo_image(
            self, scaled_size
        ) -> "ImageTk.PhotoImage":
            if scaled_size in self._scaled_light_photo_images:
                return self._scaled_light_photo_images[scaled_size]
            else:
                self._scaled_light_photo_images[
                    scaled_size
                ] = ImageTk.PhotoImage(
                    self._light_image.resize(scaled_size),
                    master=self._tk_master,
                )
                return self._scaled_light_photo_images[scaled_size]

        def _get_scaled_dark_photo_image(
            self, scaled_size
        ) -> "ImageTk.PhotoImage":
            if scaled_size in self._scaled_dark_photo_images:
                return self._scaled_dark_photo_images[scaled_size]
            else:
                self._scaled_dark_photo_images[
                    scaled_size
                ] = ImageTk.PhotoImage(
                    self._dark_image.resize(scaled_size), master=self._tk_master
                )
                return self._scaled_dark_photo_images[scaled_size]


def ctk_image_loader(source_type, source, tk_master):
    if _use_fixed_image_class:
        return CTKImageFix(Image.open(source), tk_master=tk_master)
    return CTkImage(Image.open(source))


int_properties = {
    "border_width": {},
    "border_spacing": {},
    "button_corner_radius": {},
    "button_length": {},
    "corner_radius": {},
    "height": {},
    "number_of_steps": {},
    "width": {},
    "checkbox_width": {},
    "checkbox_height": {},
    "radiobutton_width": {},
    "radiobutton_height": {},
    "border_width_unchecked": {},
    "border_width_checked": {},
    "switch_width": {},
    "switch_height": {},
    "minimum_pixel_length": {},
    "determinate_speed": {"editor": "realnumber"},
    "indeterminate_speed": {"editor": "realnumber"},
    # "": {},
}

bool_properties = {
    "dynamic_resizing": {},
    "hover": {},
    "activate_scrollbars": {},
    "round_width_to_even_numbers": {},
    "round_height_to_even_numbers": {},
    # "": {},
}

font_properties = {
    "font": {},
    "dropdown_text_font": {},
    "dropdown_font": {},
    "label_font": {},
    # "": {},
}


class CTkBaseMixin:
    uses_ctk_font = False
    uses_ctk_image = False

    def _can_set_tcl_widget_name(self) -> bool:
        """Returns True if widget accepts the tcl "name" init argument."""
        return False

    def _process_property_value(self, pname, value):
        if pname in bool_properties:
            return tk.getboolean(value)
        if pname in int_properties:
            return float(value)
        if pname in font_properties:
            fdesc = tkfontstr_to_dict(value)
            _modifiers = (
                [] if fdesc["modifiers"] is None else fdesc["modifiers"]
            )
            family = fdesc["family"]
            size = None if fdesc["size"] is None else int(fdesc["size"])
            weight = "bold" if "bold" in _modifiers else None
            slant = "italic" if "italic" in _modifiers else "roman"
            underline = True if "underline" in _modifiers else False
            overstrike = True if "overstrike" in _modifiers else False
            _font = CTkFont(family, size, weight, slant, underline, overstrike)
            return _font
        if pname == "image":
            name = Path(value).name
            if not StockImage.is_registered(name):
                StockImage.find_and_register(name)
            img = StockImage.get(value, ctk_image_loader)
            return img
        return super()._process_property_value(pname, value)

    #
    # Code generation methods
    #

    def _code_process_property_value(self, targetid, pname, value: str):
        if (
            pname in int_properties
            or pname in bool_properties
            or pname in font_properties
        ):
            return super()._process_property_value(pname, value)
        return super()._code_process_property_value(targetid, pname, value)

    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname in font_properties:
            CTkBaseMixin.uses_ctk_font = True
            fdesc = tkfontstr_to_dict(value)
            _modifiers = (
                [] if fdesc["modifiers"] is None else fdesc["modifiers"]
            )
            family = f'"{fdesc["family"]}"'
            size = None if fdesc["size"] is None else int(fdesc["size"])
            weight = '"bold"' if "bold" in _modifiers else None
            slant = '"italic"' if "italic" in _modifiers else '"roman"'
            underline = True if "underline" in _modifiers else False
            overstrike = True if "overstrike" in _modifiers else False

            fvalue = f"CTkFont({family}, {size}, {weight}, {slant}, {underline}, {overstrike})"
            code_bag[pname] = fvalue
        elif pname == "image":
            CTkBaseMixin.uses_ctk_image = True
            lines = [
                f'_img = Image.open("{value}")',
                f"{targetid}.configure(image=CTkImage(_img))",
            ]
            code_bag[pname] = lines
        else:
            super()._code_set_property(targetid, pname, value, code_bag)

    def code_imports(self):
        imports = [
            ("customtkinter", self.class_.__name__),
        ]
        if CTkBaseMixin.uses_ctk_font:
            imports.append(("customtkinter", "CTkFont"))
        if CTkBaseMixin.uses_ctk_image:
            imports.append(("customtkinter", "CTkImage"))
            imports.append(("PIL", "Image"))
        return imports


# I will register here all common properties used by customtkinter widgets
# so I don't have to repeat for each one (notice the .* at end of _builder_uid):
_builder_uid = f"{_plugin_uid}.*"

color_properties = {
    "bg_color": {
        "help": _("Color behind the widget if it has rounded corners.")
    },
    "border_color": {},
    "button_color": {},
    "button_hover_color": {},
    "checkmark_color": {},
    "dropdown_color": {},
    "dropdown_hover_color": {},
    "dropdown_text_color": {},
    "dropdown_fg_color": {},
    "fg_color": {"help": _("Main color of the widget.")},
    "hover_color": {},
    "placeholder_text_color": {},
    "progress_color": {},
    "segmented_button_fg_color": {},
    "segmented_button_selected_color": {},
    "segmented_button_selected_hover_color": {},
    "segmented_button_unselected_color": {},
    "segmented_button_unselected_hover_color": {},
    "text_color": {},
    "text_color_disabled": {},
    "scrollbar_button_color": {},
    "scrollbar_button_hover_color": {},
    "scrollbar_fg_color": {},
    "background_corner_colors": {},
    "label_fg_color": {},
    "label_text_color": {},
    # "": {},
}

for color in color_properties:
    register_custom_property(
        _builder_uid, color, "colorentry", **color_properties[color]
    )

for prop in int_properties:
    editor = int_properties[prop].pop("editor", "naturalnumber")
    register_custom_property(_builder_uid, prop, editor, **int_properties[prop])

for prop in bool_properties:
    register_custom_property(
        _builder_uid,
        prop,
        "choice",
        values=("", "True", "False"),
        state="readonly",
        **bool_properties[prop],
    )

for prop in font_properties:
    register_custom_property(_builder_uid, prop, "fontentry")

register_custom_property(_builder_uid, "placeholder_text", "entry")
register_custom_property(_builder_uid, "command", "simplecommandentry")
register_custom_property(_builder_uid, "text", "text")
register_custom_property(_builder_uid, "values", "entry")
register_custom_property(_builder_uid, "variable", "tkvarentry")
register_custom_property(_builder_uid, "label_text", "text")


register_custom_property(
    _builder_uid,
    "orientation",
    "choice",
    values=("vertical", "horizontal"),
    state="readonly",
    default_value="horizontal",
)
register_custom_property(
    _builder_uid,
    "state",
    "choice",
    values=("", "normal", "active", "disabled"),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "appearance_mode",
    "choice",
    values=("", "dark", "light"),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "color_theme",
    "choice",
    values=("", "blue", "green", "dark-blue", "sweetkind"),
    state="readonly",
    help=_("Default color theme."),
)



from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.ttk.ttkstdwidgets import TTKEntry
from tkcalendar import DateEntry

from ..tkcalendar import _designer_tab_label, _plugin_uid
from .basecalendar import CalendarBaseBO, _base_prop_desc


class DateEntryBO(CalendarBaseBO, TTKEntry):
    class_ = DateEntry
    OPTIONS_CUSTOM = CalendarBaseBO.OPTIONS_CUSTOM + ("calendar_cursor",)
    properties = (
        TTKEntry.OPTIONS_STANDARD + TTKEntry.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )
    ro_properties = CalendarBaseBO.ro_properties + ("calendar_cursor",)
    virtual_events = ("<<DateEntrySelected>>",)


_builder_id = f"{_plugin_uid}.DateEntry"

register_widget(
    _builder_id,
    DateEntryBO,
    "DateEntry",
    ("ttk", _designer_tab_label),
)

for pname, editor, options in _base_prop_desc:
    register_custom_property(_builder_id, pname, editor, **options)

register_custom_property(
    _builder_id,
    "calendar_cursor",
    "cursorentry",
)



import tkinter as tk
from pygubu.i18n import _
from pygubu.api.v1 import register_widget, register_custom_property
from pygubu.plugins.ttk.ttkstdwidgets import TTKFrame
from ttkwidgets.scaleentry import ScaleEntry

from ..ttkwidgets import _designer_tab_label, _plugin_uid


class ScaleEntryBO(TTKFrame):
    class_ = ScaleEntry
    container = False
    OPTIONS_CUSTOM = (
        "scalewidth",
        "entrywidth",
        "from_",
        "to",
        "orient",
        "compound",
        "entryscalepad",
    )
    properties = (
        TTKFrame.OPTIONS_STANDARD + TTKFrame.OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    )

    def _process_property_value(self, pname, value):
        final_value = value
        if pname in (
            "scalewidth",
            "entrywidth",
            "from_",
            "to",
            "entryscalepad",
        ):
            final_value = int(value)
        elif pname == "compound":
            final_value = tk.RIGHT
            values = {
                "top": tk.TOP,
                "bottom": tk.BOTTOM,
                "left": tk.LEFT,
                "right": tk.RIGHT,
            }
            if value in values:
                final_value = values[value]
        else:
            final_value = super(ScaleEntryBO, self)._process_property_value(
                pname, value
            )
        return final_value


_builder_uid = f"{_plugin_uid}.ScaleEntry"
register_widget(
    _builder_uid,
    ScaleEntryBO,
    "ScaleEntry",
    ("ttk", _designer_tab_label),
    group=3,
)

register_custom_property(
    _builder_uid,
    "compound",
    "choice",
    default_value=tk.RIGHT,
    help=_("side the Entry must be on."),
    values=("", tk.TOP, tk.BOTTOM, tk.LEFT, tk.RIGHT),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "orient",
    "choice",
    default_value=tk.HORIZONTAL,
    help=_("scale orientation"),
    values=("", tk.HORIZONTAL, tk.VERTICAL),
    state="readonly",
)
register_custom_property(
    _builder_uid,
    "scalewidth",
    "naturalnumber",
    help=_("width of the Scale in pixels"),
)
register_custom_property(
    _builder_uid,
    "entrywidth",
    "naturalnumber",
    help=_("width of the Entry in characters"),
)
register_custom_property(
    _builder_uid, "from_", "naturalnumber", help=_("start value of the scale")
)
register_custom_property(
    _builder_uid, "to", "naturalnumber", help=_("end value of the scale")
)
register_custom_property(
    _builder_uid,
    "entryscalepad",
    "naturalnumber",
    help=_("space between the entry and the scale"),
)



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk

import fixpath
import support
import pygubu


class TestUIDefinition_1_3(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_uidefinition_1_3.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("fmain")

    def tearDown(self):
        support.root_withdraw()

    def test_is_object_named(self):
        self.assertIsInstance(self.widget, ttk.Frame)
        is_named = self.builder.objects["fmain"].wmeta.is_named
        self.assertTrue(is_named)
        self.widget.destroy()



from ..base import Constraint, ConstraintValidator
from .notblank import NotBlank
from pygubu.i18n import _


class Required(NotBlank):
    code = "required"
    message = _("This value is required.")


class Form(Constraint):
    def get_targets(self):
        return self.CLASS_CONSTRAINT

    def validated_by(self):
        return FormValidator


class FormValidator(ConstraintValidator):
    def validate(self, form, constraint):
        for name, field in form.fields.items():
            field_constraints = []
            if field.required:
                field_constraints.append(Required())
            field_constraints.extend(field.constraints)
            for fc in field_constraints:
                validator_class = fc.validated_by()
                validator = validator_class()
                validator.initialize(self.context)
                validator.validate(field.data, fc)
            violations = self.context.violations
            if violations:
                form.add_error(name, violations)
            self.context.clear()



# encoding: utf-8
from tkinter.scrolledtext import ScrolledText

from pygubu.api.v1 import register_widget
from pygubu.plugins.tk.tkstdwidgets import TKText
from pygubu.i18n import _


class TkinterScrolledTextBO(TKText):
    class_ = ScrolledText


register_widget(
    "pygubu.builder.widgets.tkinterscrolledtext",
    TkinterScrolledTextBO,
    "tk.ScrolledText",
    (_("Control & Display"), "tk"),
)



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk

import fixpath
import pygubu
import support


class TestEntry(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = """<?xml version="1.0" ?>
<interface>
  <object class="ttk.Frame" id="mainwindow">
    <property name="height">250</property>
    <property name="width">250</property>
    <layout>
      <property name="column">0</property>
      <property name="sticky">nsew</property>
      <property name="propagate">True</property>
      <property name="row">0</property>
    </layout>
    <child>
      <object class="ttk.Entry" id="entry">
        <property name="justify">center</property>
        <property name="style">MyEntryStyle.TEntry</property>
        <property name="textvariable">entry_var</property>
        <property name="validate">key</property>
        <property name="validatecommand">entry_validate</property>
        <property name="text">Hello</property>
        <property name="validatecommand_args">%d %P</property>
        <property name="invalidcommand">entry_invalid</property>
        <property name="invalidcommand_args">%P</property>
        <layout>
          <property name="column">0</property>
          <property name="propagate">True</property>
          <property name="row">0</property>
        </layout>
      </object>
    </child>
  </object>
</interface>
"""
        self.builder = builder = pygubu.Builder()
        builder.add_from_string(xmldata)
        self.widget = builder.get_object("entry")

        self.is_style_setup = False
        if self.is_style_setup:
            self.is_style_setup = True
            s = ttk.Style()
            s.configure("MyEntryStyle.TEntry", color="Blue")

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        self.assertIsInstance(self.widget, ttk.Entry)
        self.widget.destroy()

    #    def test_class_(self):
    #        tclobj = self.widget.cget('class')
    #        class_ = str(tclobj)
    #        self.assertEqual('MyEntry', class_)
    #        self.widget.destroy()

    def test_style(self):
        stylename = "MyEntryStyle.TEntry"
        tclobj = self.widget.cget("style")
        style = str(tclobj)
        self.assertEqual(stylename, style)
        self.widget.destroy()

    def test_text(self):
        txt = self.widget.get()
        self.assertEqual("Hello", txt)
        self.widget.destroy()

    def test_variable(self):
        varname = "entry_var"
        var = self.builder.get_variable(varname)
        self.assertIsInstance(var, tk.StringVar)
        self.assertEqual("Hello", var.get())

        newlabel = "Changed"
        var.set(newlabel)
        self.assertEqual(newlabel, self.widget.get())
        self.widget.destroy()

    def test_validate_command(self):
        valid_values = ("Valid value1", "Valid value2")

        def entry_validate(action, newvalue):
            valid = False
            if action == "1":  # 1: insert 0: delete
                if newvalue in valid_values:
                    valid = True
            else:
                valid = True
            return valid

        def entry_invalid(newvalue):
            pass

        callback = {
            "entry_validate": entry_validate,
            "entry_invalid": entry_invalid,
        }

        self.builder.connect_callbacks(callback)

        self.widget.delete("0", tk.END)
        self.assertEqual("", self.widget.get())

        self.widget.insert("0", valid_values[0])
        self.assertEqual(valid_values[0], self.widget.get())

        self.widget.delete("0", tk.END)
        self.widget.insert("0", "Invalid value")
        self.assertEqual("", self.widget.get())

        self.widget.destroy()

    def test_invalid_command(self):
        invalid_text = []

        def entry_validate(action, newvalue):
            valid = False
            if action == "1":  # 1: insert 0: delete
                if newvalue == "Allowed":
                    valid = True
            else:
                valid = True
            return valid

        def entry_invalid(newvalue):
            invalid_text.append(newvalue)

        callback = {
            "entry_validate": entry_validate,
            "entry_invalid": entry_invalid,
        }

        self.widget.delete("0", tk.END)

        self.builder.connect_callbacks(callback)

        self.widget.insert("0", "Not Allowed")
        self.assertEqual(invalid_text[0], "Not Allowed")

        self.widget.destroy()

    def test_justify(self):
        prop = "justify"
        expected_value = "center"
        tclobj = self.widget.cget(prop)
        value = str(tclobj)
        self.assertEqual(expected_value, value)
        self.widget.destroy()

    def test_validate(self):
        prop = "validate"
        expected_value = "key"
        tclobj = self.widget.cget(prop)
        value = str(tclobj)
        self.assertEqual(expected_value, value)
        self.widget.destroy()


if __name__ == "__main__":
    unittest.main()



# encoding: utf-8
import tkinter as tk

from pygubu.binding import (
    ApplicationLevelBindManager as BindManager,
    remove_binding,
)


class ScrolledFrameBase:
    VERTICAL = "vertical"
    HORIZONTAL = "horizontal"
    BOTH = "both"
    _framecls = tk.Frame
    _sbarcls = tk.Scrollbar

    def __init__(self, master=None, **kw):
        self.scrolltype = kw.pop("scrolltype", self.VERTICAL)
        self.usemousewheel = tk.getboolean(kw.pop("usemousewheel", False))
        self._bindingids = []

        super().__init__(master, **kw)
        # self._framecls.__init__(self, master, **kw)

        self._container = self._framecls(self, width=200, height=200)
        self._clipper = self._framecls(self._container, width=200, height=200)
        self.innerframe = self._framecls(self._clipper)
        self.vsb = self._sbarcls(self._container)
        self.hsb = self._sbarcls(self._container, orient="horizontal")

        # variables
        self.hsbOn = 0
        self.vsbOn = 0
        self.hsbNeeded = 0
        self.vsbNeeded = 0
        self._jfraction = 0.05
        self._scrollTimer = None
        self._scrollRecurse = 0
        self._startX = 0
        self._startY = 0

        # configure scroll
        self.hsb.set(0.0, 1.0)
        self.vsb.set(0.0, 1.0)
        self.vsb.config(command=self.yview)
        self.hsb.config(command=self.xview)

        # grid
        self._container.pack(expand=True, fill="both")
        self._clipper.grid(row=0, column=0, sticky=tk.NSEW)
        # self.vsb.grid(row=0, column=1, sticky=tk.NS)
        # self.hsb.grid(row=1, column=0, sticky=tk.EW)

        self._container.rowconfigure(0, weight=1)
        self._container.columnconfigure(0, weight=1)

        # Whenever the clipping window or scrolled frame change size,
        # update the scrollbars.
        self.innerframe.bind("<Configure>", self._reposition)
        self._clipper.bind("<Configure>", self._reposition)
        self.bind("<Configure>", self._reposition)
        self._configure_mousewheel()

    # Set timer to call real reposition method, so that it is not
    # called multiple times when many things are reconfigured at the
    # same time.
    def reposition(self):
        if self._scrollTimer is None:
            self._scrollTimer = self.after_idle(self._scrollBothNow)

    # Called when the user clicks in the horizontal scrollbar.
    # Calculates new position of frame then calls reposition() to
    # update the frame and the scrollbar.
    def xview(self, mode=None, value=None, units=None):
        if isinstance(value, str):
            value = float(value)
        if mode is None:
            return self.hsb.get()
        elif mode == "moveto":
            frameWidth = self.innerframe.winfo_reqwidth()
            self._startX = value * float(frameWidth)
        else:  # mode == 'scroll'
            clipperWidth = self._clipper.winfo_width()
            if units == "units":
                jump = int(clipperWidth * self._jfraction)
            else:
                jump = clipperWidth
            self._startX = self._startX + value * jump

        self.reposition()

    # Called when the user clicks in the vertical scrollbar.
    # Calculates new position of frame then calls reposition() to
    # update the frame and the scrollbar.
    def yview(self, mode=None, value=None, units=None):

        if isinstance(value, str):
            value = float(value)
        if mode is None:
            return self.vsb.get()
        elif mode == "moveto":
            frameHeight = self.innerframe.winfo_reqheight()
            self._startY = value * float(frameHeight)
        else:  # mode == 'scroll'
            clipperHeight = self._clipper.winfo_height()
            if units == "units":
                jump = int(clipperHeight * self._jfraction)
            else:
                jump = clipperHeight
            self._startY = self._startY + value * jump

        self.reposition()

    def _reposition(self, event):
        self.reposition()

    def _getxview(self):

        # Horizontal dimension.
        clipperWidth = self._clipper.winfo_width()
        frameWidth = self.innerframe.winfo_reqwidth()
        if frameWidth <= clipperWidth:
            # The scrolled frame is smaller than the clipping window.

            self._startX = 0
            endScrollX = 1.0
            # use expand by default
            relwidth = 1
        else:
            # The scrolled frame is larger than the clipping window.
            # use expand by default
            if self._startX + clipperWidth > frameWidth:
                self._startX = frameWidth - clipperWidth
                endScrollX = 1.0
            else:
                if self._startX < 0:
                    self._startX = 0
                endScrollX = (self._startX + clipperWidth) / float(frameWidth)
            relwidth = ""

        # Position frame relative to clipper.
        self.innerframe.place(x=-self._startX, relwidth=relwidth)
        return (self._startX / float(frameWidth), endScrollX)

    def _getyview(self):

        # Vertical dimension.
        clipperHeight = self._clipper.winfo_height()
        frameHeight = self.innerframe.winfo_reqheight()
        if frameHeight <= clipperHeight:
            # The scrolled frame is smaller than the clipping window.

            self._startY = 0
            endScrollY = 1.0
            # use expand by default
            relheight = 1
        else:
            # The scrolled frame is larger than the clipping window.
            # use expand by default
            if self._startY + clipperHeight > frameHeight:
                self._startY = frameHeight - clipperHeight
                endScrollY = 1.0
            else:
                if self._startY < 0:
                    self._startY = 0
                endScrollY = (self._startY + clipperHeight) / float(frameHeight)
            relheight = ""

        # Position frame relative to clipper.
        self.innerframe.place(y=-self._startY, relheight=relheight)
        return (self._startY / float(frameHeight), endScrollY)

    # According to the relative geometries of the frame and the
    # clipper, reposition the frame within the clipper and reset the
    # scrollbars.
    def _scrollBothNow(self):
        self._scrollTimer = None

        # Call update_idletasks to make sure that the containing frame
        # has been resized before we attempt to set the scrollbars.
        # Otherwise the scrollbars may be mapped/unmapped continuously.
        self._scrollRecurse = self._scrollRecurse + 1
        self.update_idletasks()
        self._scrollRecurse = self._scrollRecurse - 1
        if self._scrollRecurse != 0:
            return

        xview = self._getxview()
        yview = self._getyview()
        self.hsb.set(xview[0], xview[1])
        self.vsb.set(yview[0], yview[1])

        require_hsb = self.scrolltype in (self.BOTH, self.HORIZONTAL)
        self.hsbNeeded = (xview != (0.0, 1.0)) and require_hsb
        require_vsb = self.scrolltype in (self.BOTH, self.VERTICAL)
        self.vsbNeeded = (yview != (0.0, 1.0)) and require_vsb

        # If both horizontal and vertical scrollmodes are dynamic and
        # currently only one scrollbar is mapped and both should be
        # toggled, then unmap the mapped scrollbar.  This prevents a
        # continuous mapping and unmapping of the scrollbars.
        if (
            self.hsbNeeded != self.hsbOn
            and self.vsbNeeded != self.vsbOn
            and self.vsbOn != self.hsbOn
        ):
            if self.hsbOn:
                self._toggleHorizScrollbar()
            else:
                self._toggleVertScrollbar()
            return

        if self.hsbNeeded != self.hsbOn:
            self._toggleHorizScrollbar()

        if self.vsbNeeded != self.vsbOn:
            self._toggleVertScrollbar()

    def _toggleHorizScrollbar(self):

        self.hsbOn = not self.hsbOn

        # interior = self #.origInterior
        if self.hsbOn:
            self.hsb.grid(row=1, column=0, sticky=tk.EW)
            # interior.grid_rowconfigure(3, minsize = self['scrollmargin'])
        else:
            self.hsb.grid_forget()
            # interior.grid_rowconfigure(3, minsize = 0)

    def _toggleVertScrollbar(self):

        self.vsbOn = not self.vsbOn

        # interior = self#.origInterior
        if self.vsbOn:
            self.vsb.grid(row=0, column=1, sticky=tk.NS)
            # interior.grid_columnconfigure(3, minsize = self['scrollmargin'])
        else:
            self.vsb.grid_forget()
            # interior.grid_columnconfigure(3, minsize = 0)

    def configure(self, cnf=None, **kw):
        if cnf:
            return super().configure(cnf, **kw)
        key = "usemousewheel"
        if key in kw:
            self.usemousewheel = tk.getboolean(kw.pop(key))
            self._configure_mousewheel()
        return self._framecls.configure(self, cnf, **kw)

    config = configure

    def cget(self, key):
        option = "usemousewheel"
        if key == option:
            return self.usemousewheel
        # return super(ScrolledFrameBase, self).cget(key)
        return self._framecls.cget(self, key)

    __getitem__ = cget

    def _configure_mousewheel(self):
        if self.usemousewheel:
            BindManager.init_mousewheel_binding(self)

            if self.hsb and not hasattr(self.hsb, "on_mousewheel"):
                self.hsb.on_mousewheel = BindManager.make_onmousewheel_cb(
                    self, "x", 2
                )
            if self.vsb and not hasattr(self.vsb, "on_mousewheel"):
                self.vsb.on_mousewheel = BindManager.make_onmousewheel_cb(
                    self, "y", 2
                )

            main_sb = self.vsb or self.hsb
            if main_sb:
                self.on_mousewheel = main_sb.on_mousewheel
                BindManager.mousewheel_bind(self)
            for s in (self.vsb, self.hsb):
                if s:
                    BindManager.mousewheel_bind(s)
        else:
            for w in (self, self.vsb, self.hsb):
                BindManager.mousewheel_unbind(w)


class TkScrolledFrame(ScrolledFrameBase, tk.Frame):
    ...



# encoding: utf-8
from pygubu.api.v1 import register_widget
from pygubu.plugins.ttk.ttkstdwidgets import (
    TTKTreeviewBO,
    TTKTreeviewColumnBO,
)
from .scrollbarhelper import TTKSBHelperBO
from pygubu.i18n import _
from pygubu.widgets.filterabletreeview import FilterableTreeview


class FilterableTreeviewBO(TTKTreeviewBO):
    class_ = FilterableTreeview
    # virtual_events = TTKTreeviewBO.virtual_events + (
    #     "<<FilterableTreeview:NoResults>>",
    # )


_builder_uid = "pygubu.widgets.FilterableTreeview"
if _builder_uid not in TTKTreeviewColumnBO.allowed_parents:
    TTKTreeviewColumnBO.add_allowed_parent(_builder_uid)
if _builder_uid not in TTKSBHelperBO.allowed_children:
    TTKSBHelperBO.add_allowed_child(_builder_uid)

register_widget(
    _builder_uid,
    FilterableTreeviewBO,
    "FilterableTreeview",
    (_("Pygubu Widgets"), "ttk"),
)



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk


import fixpath
import pygubu
import support


expected_text = """In the designer:
- Add a text widget
- Set state to disabled
- Enter content in the text property
- Text content should be set by pygubu-designer.
"""


class TestText(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_text_issue58.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.widget = builder.get_object("mainframe")
        self.text = builder.get_object("Text_1")

    def tearDown(self):
        support.root_withdraw()

    def test_set_text_in_disabled_status(self):
        text_content = self.text.get("0.0", tk.END)
        self.assertEqual(text_content, expected_text)
        self.widget.destroy()



# encoding: utf-8
import os
import sys
import unittest
import tkinter as tk
import tkinter.ttk as ttk


import fixpath
import pygubu
import support
from pygubu.widgets.pathchooserinput import PathChooserInput


class TestPathChooserInput(unittest.TestCase):
    def setUp(self):
        support.root_deiconify()
        xmldata = "test_pathchooserinput.ui"
        self.builder = builder = pygubu.Builder()
        builder.add_from_file(xmldata)
        self.mainwindow = builder.get_object("mainwindow")
        self.widget = builder.get_object("pathchooserinput1")

    def tearDown(self):
        support.root_withdraw()

    def test_class(self):
        self.assertIsInstance(self.widget, PathChooserInput)
        self.widget.destroy()

    def test_path_changed_event(self):
        success = []

        def on_path_changed(event=None):
            success.append(1)

        bag = {"on_path_changed": on_path_changed}
        self.builder.connect_callbacks(bag)

        self.widget.configure(path="/new/path")
        self.assertTrue(success)
        self.widget.destroy()

    def test_type(self):
        itype = str(self.widget.cget("type"))
        self.assertEqual("file", itype)
        self.widget.destroy()

    def test_path(self):
        path = str(self.widget.cget("path"))
        self.assertEqual("/home/user", path)
        self.widget.destroy()

    def test_path_dictionary_like(self):
        path = str(self.widget["path"])
        self.assertEqual("/home/user", path)
        self.widget.destroy()

    def test_state(self):
        # allowed states normal/disabled/readonly

        # normal
        state = str(self.widget.cget("state"))
        self.assertEqual("normal", state)

        # disabled
        self.widget.config(state="disabled")
        state = str(self.widget.cget("state"))
        self.assertEqual("disabled", state)

        # readonly
        self.widget.config(state="readonly")
        state = str(self.widget.cget("state"))
        self.assertEqual("readonly", state)

        self.widget.destroy()



import tkinter as tk
import tkinter.ttk as ttk
from typing import Callable, Optional


_filter_func = Callable[[ttk.Treeview, str, str], bool]


class BasicFilter:
    def __call__(self, tree, itemid, filter_value: str) -> bool:
        # Default filter match function.
        txt = tree.item(itemid, "text").lower()
        match_found = filter_value in txt
        return match_found


class FilterableTreeview(ttk.Treeview):
    def __init__(self, *args, filter_func: Optional[_filter_func] = None, **kw):
        super().__init__(*args, **kw)
        self.filter_active = False
        self.filter_value = ""
        self.filter_prev_value = None
        self.filter_prev_selitem = None
        self._detached = []
        self._expanded = set()
        self.filter_func = BasicFilter() if filter_func is None else filter_func

    #
    # Filter functions
    #
    def _see(self, item):
        # The item may have been deleted.
        if self.exists(item):
            self.see(item)

    def _see_later(self, item):
        self.after_idle(lambda: self._see(item))

    def filter_by(self, fvalue):
        """Filters treeview"""

        self.filter_remove()
        if not fvalue:
            return

        self.filter_value = fvalue
        # self._expand_all()
        self.selection_set("")

        children = self.get_children("")
        for item in children:
            match_, detached = self._filter_and_detach(item)
            if detached:
                self._detached.extend(detached)
            if match_:
                self.item(item, open=True)
        for i, p, idx in self._detached:
            self.detach(i)
        self.filter_active = True

    def filter_remove(self, remember=False):
        """Removes filter and reattaches hidden items.
        When filter is applied you can not traverse through detached items.
        So to be able to navigate all tree items with filter active you use
        the following approach:

          tree.filter_remove(remember=True)
          # Do operation, navigate tree items.
          tree.filter_restore()
        """
        if self.filter_active:
            sitem = None
            selection = self.selection()
            if selection:
                sitem = selection[0]
                self._see_later(sitem)
            if remember:
                self.filter_prev_value = self.filter_value
                self.filter_prev_selitem = sitem
            self._reattach()
            self._restore_open_state()
            self.filter_value = ""
        self.filter_active = False

    def filter_restore(self):
        if self.filter_prev_value:
            self.filter_value = self.filter_prev_value
            item = self.filter_prev_selitem
            if item and self.exists(item):
                self.selection_set(item)
                self._see_later(item)
            # clear
            self.filter_prev_value = ""
            self.filter_prev_selitem = None
            # Re apply filter?
            self.filter_by(self.filter_value)

    def expand_to(self, target_item, start_item=""):
        """Search and expand tree to see itemid."""
        self._expand_to(start_item, target_item)

    def _expand_to(self, rootitem, target_item):
        found = rootitem == target_item
        if not found:
            children = self.get_children(rootitem)
            found = target_item in children
            if not found:
                for child_item in children:
                    found = self._expand_to(child_item, target_item)
                    if found:
                        break
        if found and rootitem != "":
            self.item(rootitem, open=True)
        return found

    def expand_all(self, rootitem=""):
        children = self.get_children(rootitem)
        for item in children:
            self._expand_all(item)
        if rootitem != "" and children:
            self.item(rootitem, open=True)

    def _restore_open_state(self, rootitem=""):
        if rootitem != "":
            was_open = rootitem in self._expanded
            self.item(rootitem, open=was_open)
        children = self.get_children(rootitem)
        for item in children:
            self._restore_open_state(item)
        if rootitem == "":
            # End of recursion clear expanded set
            self._expanded.clear()

    def _reattach(self):
        """Reinsert the hidden items."""
        for item, p, idx in self._detached:
            # The item may have been deleted.
            if self.exists(item) and self.exists(p):
                self.move(item, p, idx)
        self._detached = []

    def _filter_and_detach(self, item) -> (bool, list):
        """Hide items from treeview that do not match the search string."""
        to_detach = []
        children_det = []
        children_match = False
        if self.item(item, "open"):
            self._expanded.add(item)
        match_found = self.filter_func(self, item, self.filter_value)
        parent = self.parent(item)
        idx = self.index(item)
        children = self.get_children(item)
        if children:
            for child in children:
                match, detach = self._filter_and_detach(child)
                children_match = children_match | match
                if detach:
                    children_det.extend(detach)
        if match_found:
            self.item(item, open=True)
            if children_det:
                to_detach.extend(children_det)
        else:
            if children_match:
                self.item(item, open=True)
                if children_det:
                    to_detach.extend(children_det)
            else:
                to_detach.append((item, parent, idx))
        match_found = match_found | children_match
        return match_found, to_detach

    #
    # End Filter functions
    #



from pygubu.api.v1 import BuilderLoaderPlugin


class StandardTKWidgetsLoader(BuilderLoaderPlugin):
    _module = "pygubu.plugins.tk.tkstdwidgets"

    def do_activate(self) -> bool:
        return True

    def get_module_for(self, identifier: str) -> str:
        return self._module

    def get_all_modules(self):
        return (self._module,)

    def can_load(self, identifier: str) -> bool:
        return identifier.startswith("tk.")



import customtkinter as ctk
import pygubu.plugins.customtkinter.tabview
import pygubu.plugins.customtkinter.scrollableframe

from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from pygubu.i18n import _
from pygubu.utils.datatrans import ListDTO
from pygubu.plugins.tk.tkstdwidgets import TKCanvas as TKCanvasBO

from ..customtkinter import _designer_tab_label, _plugin_uid
from .ctkbase import (
    CTkBaseMixin,
    GCONTAINER,
    GDISPLAY,
    GINPUT,
)


class CTkFrameBO(CTkBaseMixin, BuilderObject):
    class_ = ctk.CTkFrame
    container = True
    properties = (
        "width",
        "height",
        "corner_radius",
        "border_width",
        "bg_color",
        "fg_color",
        "border_color",
        "background_corner_colors",
    )


_builder_uid = f"{_plugin_uid}.CTkFrame"
register_widget(
    _builder_uid,
    CTkFrameBO,
    "CTkFrame",
    ("ttk", _designer_tab_label),
    group=GCONTAINER,
)


class CTkLabelBO(CTkBaseMixin, BuilderObject):
    class_ = ctk.CTkLabel
    properties = (
        "anchor",
        "compound",
        "cursor",
        "image",
        "justify",
        "font",
        "height",
        "padx",
        "pady",
        "state",
        "takefocus",
        "text",
        "textvariable",
        "underline",
        "width",
        # CTK properties
        "corner_radius",
        "bg_color",
        "fg_color",
        "text_color",
        "text_color_disabled",
        "border_color",
        "border_width",
    )


_builder_uid = f"{_plugin_uid}.CTkLabel"
register_widget(
    _builder_uid,
    CTkLabelBO,
    "CTkLabel",
    ("ttk", _designer_tab_label),
    group=GDISPLAY,
)


class CTkProgressBarBO(CTkBaseMixin, BuilderObject):
    class_ = ctk.CTkProgressBar
    allow_bindings = False
    properties = (
        "width",
        "height",
        "variable",
        "mode",
        "orientation",
        # CTK properties
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "corner_radius",
        "progress_color",
        "determinate_speed",
        "indeterminate_speed",
    )
    ro_properties = ("orientation",)


_builder_uid = f"{_plugin_uid}.CTkProgressBar"
register_widget(
    _builder_uid,
    CTkProgressBarBO,
    "CTkProgressBar",
    ("ttk", _designer_tab_label),
    group=GDISPLAY,
)

register_custom_property(
    _builder_uid,
    "variable",
    "tkvarentry",
    type_choices=("int", "double"),
    type_default="int",
)


class CTkButtonBO(CTkBaseMixin, BuilderObject):
    class_ = ctk.CTkButton
    allow_bindings = False
    properties = (
        "text",
        "width",
        "height",
        "textvariable",
        "image",
        "compound",
        "state",
        "command",
        # CTK properties
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "border_spacing",
        "corner_radius",
        "hover_color",
        "text_color",
        "text_color_disabled",
        "hover",
        "font",
        "background_corner_colors",
        "round_width_to_even_numbers",
        "round_height_to_even_numbers",
        "anchor",
    )
    command_properties = ("command",)
    ro_properties = ("hover",)


_builder_uid = f"{_plugin_uid}.CTkButton"
register_widget(
    _builder_uid,
    CTkButtonBO,
    "CTkButton",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)


class CTkSliderBO(CTkBaseMixin, BuilderObject):
    class_ = ctk.CTkSlider
    allow_bindings = False
    properties = (
        "width",
        "height",
        "variable",
        "from_",
        "to",
        "command",
        "state",
        # CTK properties
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "corner_radius",
        "progress_color",
        "button_color",
        "button_hover_color",
        "button_corner_radius",
        "button_length",
        "number_of_steps",
        "orientation",
    )
    command_properties = ("command",)
    ro_properties = (
        "orientation",
        "button_length",
    )

    def _code_define_callback_args(self, cmd_pname, cmd):
        return ("value",)


_builder_uid = f"{_plugin_uid}.CTkSlider"
register_widget(
    _builder_uid,
    CTkSliderBO,
    "CTkSlider",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)

register_custom_property(
    _builder_uid,
    "variable",
    "tkvarentry",
    type_choices=("int", "double"),
    type_default="int",
)


class CTkEntryBO(CTkBaseMixin, BuilderObject):
    class_ = ctk.CTkEntry
    allow_bindings = False
    properties = (
        "borderwidth",
        "cursor",
        "exportselection",
        "font",
        "insertborderwidth",
        "insertofftime",
        "insertontime",
        "insertwidth",
        "justify",
        "selectborderwidth",
        "takefocus",
        "textvariable",
        "xscrollcommand",
        # specific
        "invalidcommand",
        "readonlybackground",
        "show",
        "state",
        "validate",
        "validatecommand",
        "width",
        # custom
        "text",
        # CTK options
        "bg_color",
        "fg_color",
        "border_color",
        "border_width",
        "corner_radius",
        "text_color",
        "placeholder_text_color",
        "placeholder_text",
    )

    def _set_property(self, target_widget, pname, value):
        if pname == "text":
            super()._set_property(target_widget._entry, pname, value)
        else:
            super()._set_property(target_widget, pname, value)


_builder_uid = f"{_plugin_uid}.CTkEntry"
register_widget(
    _builder_uid,
    CTkEntryBO,
    "CTkEntry",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)


_list_dto = ListDTO()


class CTkOptionMenuBO(CTkBaseMixin, BuilderObject):
    class_ = ctk.CTkOptionMenu
    allow_bindings = False
    properties = (
        "command",
        "variable",
        "values",
        "bg_color",
        "fg_color",
        "button_color",
        "button_hover_color",
        "text_color",
        "text_color_disabled",
        "dropdown_hover_color",
        "dropdown_text_color",
        "dropdown_color",
        "dropdown_font",
        "width",
        "height",
        "corner_radius",
        "state",
        "dynamic_resizing",
        "font",
    )
    command_properties = ("command",)

    def _process_property_value(self, pname, value):
        if pname == "values":
            return _list_dto.transform(value)
        return super()._process_property_value(pname, value)

    def _code_define_callback_args(self, cmd_pname, cmd):
        return ("current_value",)

    def _code_process_property_value(self, targetid, pname, value: str):
        if pname == "values":
            return super()._process_property_value(pname, value)
        return super()._code_process_property_value(targetid, pname, value)


_builder_uid = f"{_plugin_uid}.CTkOptionMenu"
_ctk_values_help = _(
    "Specifies the list of values to display. "
    "In code you can pass any iterable. "
    'In Designer, a json like list: ["item1", "item2"]'
)
register_widget(
    _builder_uid,
    CTkOptionMenuBO,
    "CTkOptionMenu",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)
register_custom_property(_builder_uid, "values", "entry", help=_ctk_values_help)
register_custom_property(
    _builder_uid,
    "state",
    "choice",
    values=("", "normal", "disabled", "readonly"),
    state="readonly",
)


class CTkComboBoxBO(CTkBaseMixin, BuilderObject):
    class_ = ctk.CTkComboBox
    allow_bindings = False
    properties = (
        "corner_radius",
        "border_width",
        "bg_color",
        "fg_color",
        "border_color",
        "button_color",
        "button_hover_color",
        "dropdown_fg_color",
        "dropdown_hover_color",
        "dropdown_text_color",
        "text_color",
        "text_color_disabled",
        "font",
        "dropdown_font",
        "command",
        "variable",
        "values",
        "width",
        "height",
        "state",
        "hover",
    )
    command_properties = ("command",)
    ro_properties = ("hover",)

    def _process_property_value(self, pname, value):
        if pname == "values":
            return _list_dto.transform(value)
        return super()._process_property_value(pname, value)

    def _code_define_callback_args(self, cmd_pname, cmd):
        return ("value",)

    def _code_process_property_value(self, targetid, pname, value: str):
        if pname == "values":
            return super()._process_property_value(pname, value)
        return super()._code_process_property_value(targetid, pname, value)


_builder_uid = f"{_plugin_uid}.CTkComboBox"
register_widget(
    _builder_uid,
    CTkComboBoxBO,
    "CTkComboBox",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)
register_custom_property(_builder_uid, "values", "entry", help=_ctk_values_help)
register_custom_property(
    _builder_uid,
    "state",
    "choice",
    values=("", "normal", "disabled", "readonly"),
    state="readonly",
)


class CTkCheckBoxBO(CTkBaseMixin, BuilderObject):
    class_ = ctk.CTkCheckBox
    allow_bindings = False
    properties = (
        "width",
        "height",
        "checkbox_width",
        "checkbox_height",
        "corner_radius",
        "border_width",
        "bg_color",
        "fg_color",
        "hover_color",
        "border_color",
        "checkmark_color",
        "text_color",
        "text_color_disabled",
        "text",
        "font",
        "textvariable",
        "state",
        "hover",
        "command",
        "onvalue",
        "offvalue",
        "variable",
    )
    command_properties = ("command",)
    ro_properties = ("hover", "onvalue", "offvalue")


_builder_uid = f"{_plugin_uid}.CTkCheckBox"
register_widget(
    _builder_uid,
    CTkCheckBoxBO,
    "CTkCheckBox",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)


class CTkRadioButtonBO(CTkBaseMixin, BuilderObject):
    class_ = ctk.CTkRadioButton
    allow_bindings = False
    properties = (
        "width",
        "height",
        "radiobutton_width",
        "radiobutton_height",
        "corner_radius",
        "border_width_unchecked",
        "border_width_checked",
        "bg_color",
        "fg_color",
        "hover_color",
        "border_color",
        "text_color",
        "text_color_disabled",
        "text",
        "font",
        "textvariable",
        "variable",
        "value",
        "state",
        "hover",
        "command",
    )
    command_properties = ("command",)
    ro_properties = ("hover", "value")


_builder_uid = f"{_plugin_uid}.CTkRadioButton"
register_widget(
    _builder_uid,
    CTkRadioButtonBO,
    "CTkRadioButton",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)


class CTkSwitchBO(CTkBaseMixin, BuilderObject):
    class_ = ctk.CTkSwitch
    allow_bindings = False
    properties = (
        "width",
        "height",
        "switch_width",
        "switch_height",
        "corner_radius",
        "border_width",
        "button_length",
        "bg_color",
        "fg_color",
        "border_color",
        "progress_color",
        "button_color",
        "button_hover_color",
        "text_color",
        "text_color_disabled",
        "text",
        "font",
        "textvariable",
        "onvalue",
        "offvalue",
        "variable",
        "hover",
        "command",
        "state",
    )
    command_properties = ("command",)
    ro_properties = (
        "onvalue",
        "offvalue",
        "hover",
        "text_color",
        "text_color_disabled",
    )


_builder_uid = f"{_plugin_uid}.CTkSwitch"
register_widget(
    _builder_uid,
    CTkSwitchBO,
    "CTkSwitch",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)


class CTkTextboxBO(CTkBaseMixin, BuilderObject):
    class_ = ctk.CTkTextbox
    properties = (
        "autoseparators",
        "cursor",
        "exportselection",
        "insertborderwidth",
        "insertofftime",
        "insertontime",
        "insertwidth",
        "maxundo",
        "padx",
        "pady",
        "selectborderwidth",
        "spacing1",
        "spacing2",
        "spacing3",
        "state",
        "tabs",
        "takefocus",
        "undo",
        "wrap",
        # ctk
        "width",
        "height",
        "corner_radius",
        "border_width",
        "border_spacing",
        "bg_color",
        "fg_color",
        "border_color",
        "text_color",
        "font",
        "scrollbar_button_color",
        "scrollbar_button_hover_color",
        "activate_scrollbars",
        # custom
        "text",
    )

    def _set_property(self, target_widget, pname, value):
        if pname == "text":
            target_widget = target_widget._textbox
            state = target_widget.cget("state")
            if state == ctk.DISABLED:
                target_widget.configure(state=ctk.NORMAL)
                target_widget.insert("0.0", value)
                target_widget.configure(state=ctk.DISABLED)
            else:
                target_widget.insert("0.0", value)
        else:
            super()._set_property(target_widget, pname, value)

    def _code_set_property(self, targetid, pname, value, code_bag):
        if pname == "text":
            state_value = ""
            if "state" in self.wmeta.properties:
                state_value = self.wmeta.properties["state"]
            sval = self.builder.code_translate_str(value)
            lines = [
                f"_text_ = {sval}",
            ]
            if state_value == ctk.DISABLED:
                lines.extend(
                    (
                        f'{targetid}.configure(state="normal")',
                        f'{targetid}.insert("0.0", _text_)',
                        f'{targetid}.configure(state="disabled")',
                    )
                )
            else:
                lines.append(f'{targetid}.insert("0.0", _text_)')
            code_bag[pname] = lines
        else:
            super()._code_set_property(targetid, pname, value, code_bag)


_builder_uid = f"{_plugin_uid}.CTkTextbox"
register_widget(
    _builder_uid,
    CTkTextboxBO,
    "CTkTextbox",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)

register_custom_property(
    _builder_uid,
    "wrap",
    "choice",
    values=("", ctk.CHAR, ctk.WORD, ctk.NONE),
    state="readonly",
)


class CTkScrollbarBO(CTkBaseMixin, BuilderObject):
    class_ = ctk.CTkScrollbar
    properties = (
        "width",
        "height",
        "orientation",
        "command",
        # CTK
        "cursor",
        "corner_radius",
        "border_spacing",
        "minimum_pixel_length",
        "bg_color",
        "fg_color",
        "button_color",
        "button_hover_color",
        "hover",
    )
    ro_properties = ("orientation", "cursor")
    command_properties = ("command",)


_builder_uid = f"{_plugin_uid}.CTkScrollbar"
register_widget(
    _builder_uid,
    CTkScrollbarBO,
    "CTkScrollbar",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)

register_custom_property(
    _builder_uid,
    "orientation",
    "choice",
    values=("vertical", "horizontal"),
    state="readonly",
    default_value="vertical",
)
register_custom_property(
    _builder_uid,
    "command",
    "scrollcommandentry",
)


class CTkSegmentedButtonBO(CTkBaseMixin, BuilderObject):
    class_ = ctk.CTkSegmentedButton
    allow_bindings = False
    properties = (
        "width",
        "height",
        "corner_radius",
        "border_width",
        "bg_color",
        "fg_color",
        "selected_color",
        "selected_hover_color",
        "unselected_color",
        "unselected_hover_color",
        "text_color",
        "text_color_disabled",
        "background_corner_colors",
        "font",
        "values",
        "variable",
        "dynamic_resizing",
        "command",
        "state",
    )
    command_properties = ("command",)
    ro_properties = ("hover",)

    def _process_property_value(self, pname, value):
        if pname == "values":
            return _list_dto.transform(value)
        return super()._process_property_value(pname, value)

    def _code_define_callback_args(self, cmd_pname, cmd):
        return ("current_value",)

    def _code_process_property_value(self, targetid, pname, value: str):
        if pname == "values":
            return super()._process_property_value(pname, value)
        return super()._code_process_property_value(targetid, pname, value)


_builder_uid = f"{_plugin_uid}.CTkSegmentedButton"
register_widget(
    _builder_uid,
    CTkSegmentedButtonBO,
    "CTkSegmentedButton",
    ("ttk", _designer_tab_label),
    group=GINPUT,
)

register_custom_property(_builder_uid, "values", "entry", help=_ctk_values_help)


class CTkCanvasBO(TKCanvasBO):
    class_ = ctk.CTkCanvas


_builder_uid = f"{_plugin_uid}.CTkCanvas"
register_widget(
    _builder_uid,
    CTkCanvasBO,
    "CTkCanvas",
    ("ttk", _designer_tab_label),
    group=GDISPLAY,
)



import tkinter as tk
import tkinter.ttk as ttk

from pygubu.api.v1 import (
    register_widget,
    register_custom_property,
    BuilderObject,
)
from pygubu.i18n import _

import pygubu.widgets.dockfw.widgets as widgets


class DockWidgetBaseBO(BuilderObject):
    def _get_init_args(self, extra_init_args: dict = None):
        args = super()._get_init_args(extra_init_args)
        if "uid" not in args:
            args["uid"] = self.wmeta.identifier
        return args


class DockFrameBO(DockWidgetBaseBO):
    class_ = widgets.DockFrame
    container = True
    container_layout = False
    maxchildren = 1
    virtual_events = (widgets.DockFrame.EVENT_LAYOUT_CHANGED,)


_builder_id = dockframe_uid = "pygubu.widgets.dockframe"
register_widget(
    _builder_id, DockFrameBO, "DockFrame", ("ttk", _("Pygubu Widgets"))
)


class DockPaneBO(DockWidgetBaseBO):
    class_ = widgets.DockPane
    container = True
    container_layout = False
    layout_required = False
    properties = ("orient", "weight")
    ro_properties = properties
    allowed_parents = (dockframe_uid,)

    @classmethod
    def canbe_child_of(cls, parent_builder, classname):
        allowed = False
        if parent_builder in (DockFrameBO, DockPaneBO):
            allowed = True
        return allowed

    def __init__(self, builder, wmeta):
        super().__init__(builder, wmeta)
        self.pane_widget = None

    def realize(self, parent, extra_init_args: dict = None):
        self.widget: widgets.DockFrame = parent.widget
        args = self._get_init_args(extra_init_args)
        if not self.widget.main_pane:
            args["main_pane"] = True
        pweight = args.pop("weight", 1)
        self.pane_widget = self.widget.new_pane(**args)
        if isinstance(parent, DockPaneBO):
            parent.pane_widget.add_pane(self.pane_widget, weight=pweight)
        return self.widget

    def configure(self, target=None):
        pass

    def layout(self, target=None):
        pass


_builder_id = dockpane_uid = "pygubu.widgets.dockpane"
DockPaneBO.add_allowed_child(_builder_id)

register_widget(
    _builder_id, DockPaneBO, "DockPane", ("ttk", _("Pygubu Widgets"))
)

register_custom_property(
    _builder_id,
    "weight",
    "integernumber",
    help=_("The weight value for the pane."),
)


class DockWidgetBO(DockWidgetBaseBO):
    class_ = widgets.DockWidget
    container = True
    container_layout = True
    layout_required = False
    properties = ("grouped", "weight", "title")
    ro_properties = ("grouped", "weight")
    allowed_parents = (dockframe_uid, dockpane_uid)

    @classmethod
    def canbe_child_of(cls, parent_builder, classname):
        allowed = False
        if parent_builder is DockPaneBO:
            allowed = True
        return allowed

    def _process_property_value(self, pname, value):
        if pname == "grouped":
            return tk.getboolean(value)
        return super()._process_property_value(pname, value)

    def realize(self, parent, extra_init_args: dict = None):
        dock = parent.pane_widget.maindock
        args: dict = self._get_init_args(extra_init_args)
        grouped = args.pop("grouped", False)
        weight = args.pop("weight", 1)
        self.widget = dock.new_widget(**args)
        super().configure(target=self.widget)  # hack used here
        parent.pane_widget.add_widget(
            self.widget, grouped=grouped, weight=weight
        )

    def get_child_master(self):
        return self.widget.fcenter

    def configure(self, target=None):
        pass


_builder_id = "pygubu.widgets.dockwidget"
DockPaneBO.add_allowed_child(_builder_id)

register_widget(
    _builder_id, DockWidgetBO, "DockWidget", ("ttk", _("Pygubu Widgets"))
)

register_custom_property(
    _builder_id,
    "grouped",
    "choice",
    values=("", "true", "false"),
    state="readonly",
)

register_custom_property(
    _builder_id,
    "weight",
    "integernumber",
    help=_("The weight value of the widget in the pane"),
)



from ..base import Constraint, ConstraintValidator


class IsTrue(Constraint):
    code = "not_true_error"
    message = "This value should be true."
    true_values = (True, "1", 1)

    def __init__(
        self, *args, message=None, allow_none=False, true_values=None, **kw
    ):
        super().__init__(*args, **kw)
        if message is not None:
            self.message = message
        if true_values is not None:
            self.true_values = true_values
        self.allow_none = allow_none

    def validated_by(self):
        return IsTrueValidator


class IsTrueValidator(ConstraintValidator):
    def validate(self, value, constraint):
        if constraint.allow_none and value is None:
            return
        if value not in constraint.true_values:
            self.context.add_violation(
                message=constraint.message,
                constraint=constraint,
                code=constraint.code,
                params=None,
            )



import tkinter as tk
import tkinter.ttk as ttk
import logging

from pygubu.widgets.dockfw.dockwidget import (
    DockWidgetBase,
    DockPane,
    DockWidget,
)
from pygubu.widgets.dockfw.framework import (
    DockingFramework,
    IDockFrame,
)


logger = logging.getLogger(__name__)


class DockFrame(DockWidgetBase, IDockFrame):
    """A Dock Frame, where user can configure the layout
    of the frames within a window.
    Based on https://wiki.tcl-lang.org/page/Docking+framework

    Generates <<DocFrame::LayoutChanged>> event when the user
    changes the layout.
    """

    EVENT_LAYOUT_CHANGED = "<<DocFrame::LayoutChanged>>"

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.dock_widgets: dict[str, DockWidget] = {}
        self._main_pane = None
        self._emit_cbid = None

    def indicators_visible(self, visible: bool):
        if self._main_pane:
            self._main_pane.indicators_visible(visible)

    @property
    def main_pane(self):
        return self._main_pane

    def _add_pane_to_pane(
        self, topane: DockPane, newpane: DockPane, position=tk.END, **pane_kw
    ):
        # remove newpane as child of current pane if any
        newpane.detach_from_parent()

        topane.panedw.insert(position, newpane, **pane_kw)
        newpane.parent_pane = topane
        topane.add_dwchild(newpane)

    def _add_widget_to_pane(
        self, topane, widget, grouped=False, position=tk.END, weight=1
    ):
        nb = None
        if grouped:
            panes = topane.panedw.panes()
            if panes:
                nb = self.nametowidget(panes[-1])
        if nb is None:
            nb = ttk.Notebook(self, padding=0)
            DockingFramework.init_binding(nb)
            # topane.panedw.add(nb, weight=1)
            topane.panedw.insert(position, nb, weight=weight)
        nb.add(widget, text=widget.title)
        widget.tkraise()
        widget.detach_from_parent()
        topane.add_dwchild(widget)
        # widget.parent_pane = topane
        widget.noteb = nb
        self.dock_widgets[widget.uid] = widget

    def new_pane(self, *, main_pane=False, **pane_kw) -> DockPane:
        """Creates new DockPane locked to this DockFrame."""
        pane = DockPane(self, maindock=self, **pane_kw)
        if main_pane:
            self.set_main_pane(pane)
        return pane

    def set_main_pane(self, pane):
        pane.pack(in_=self.fcenter, expand=True, fill=tk.BOTH)
        self._main_pane = pane

    def new_widget(self, **widget_kw) -> DockWidget:
        """Creates new DockWidget locked to this DockFrame."""
        widget = DockWidget(self, maindock=self, **widget_kw)
        return widget

    def _add_widget_to_group(
        self, tgroup, swidget: DockWidget, position=tk.END
    ):
        nb: ttk.Notebook = tgroup.noteb
        nb.insert(position, swidget, text=swidget.title)
        nb.select(position)
        swidget.noteb = nb
        if tgroup != swidget:
            swidget.detach_from_parent()
            tgroup.parent_pane.add_dwchild(swidget)
        # swidget.parent_pane = tgroup.parent_pane
        swidget.tkraise()
        self.dock_widgets[swidget.uid] = swidget

    def _move_into_widget_group(self, twidget, swidget, position):
        noteb = swidget.noteb
        pane = swidget.parent_pane
        self._add_widget_to_group(twidget, swidget, position)
        self._clear_notebook(noteb)
        self._clear_pane(pane)
        self._emit_layout_changed()

    def _move_into_pane_side(self, swidget, tpane, side, position=None):
        logger.debug("Move to into pane side. position %s", position)

        move_in_same_pane = swidget.parent_pane == tpane
        simple_sides = "ns" if tpane.orient == tk.VERTICAL else "we"

        noteb = swidget.noteb
        pane = swidget.parent_pane
        if position is None:
            position = 0 if side in "nw" else tk.END

        if move_in_same_pane:
            if side in simple_sides:
                # if widget is in a group, must detach and re-insert widget
                if swidget.is_grouped:
                    self._add_widget_to_pane(tpane, swidget, position=position)
                elif tpane.count > 1:
                    # pane has more than the moving widget
                    # reposition widget in same pane
                    logger.debug(
                        "Reposition widget in same pane. position %s", position
                    )
                    tpane.panedw.insert(position, noteb)
                else:
                    logger.debug("No need to move")
            else:
                self._move_into_pane_side_complex(
                    swidget, tpane, side, position
                )
        else:
            if side in simple_sides:
                self._add_widget_to_pane(tpane, swidget, position=position)
            else:
                self._move_into_pane_side_complex(
                    swidget, tpane, side, position
                )
        self._clear_notebook(noteb)
        self._clear_pane(pane)
        self._emit_layout_changed()

    def _move_into_pane_side_complex(self, swidget, tpane, side, position=None):
        move_to_new_pane = True
        # check if parent pane is in sync
        parent_pane = tpane.parent_pane
        if parent_pane:
            in_sync = side in (
                "ns" if parent_pane.orient == tk.VERTICAL else "we"
            )
            if in_sync:
                logger.debug("Parent pane in sync, adding into parent.")
                index = self._get_position_in_pane(tpane)
                logger.debug("Pane position: %s side: %s", index, side)
                if side not in "nw":
                    index += 1
                    index = tk.END if index >= parent_pane.count else index
                logger.debug("final index: %s", index)
                self._add_widget_to_pane(parent_pane, swidget, position=index)
                move_to_new_pane = False
        if move_to_new_pane:
            logger.debug("Moving to new pane.")
            self._move_to_new_pane(swidget, tpane, side)

    def _move_to_new_pane(self, swidget: DockWidget, tpane: DockPane, side):
        """Move DockWidget 'swidget' to a new pane inside target pane 'tpane'."""
        logger.debug("Move to new pane.")
        new_orient = (
            tk.HORIZONTAL if tpane.orient == tk.VERTICAL else tk.VERTICAL
        )
        main_pane = True if tpane == self._main_pane else False
        new_pane = self.new_pane(main_pane=main_pane, orient=new_orient)
        parent_pane = tpane.parent_pane
        self._add_pane_to_pane(new_pane, tpane)
        if parent_pane:
            self._add_pane_to_pane(parent_pane, new_pane)
        new_pos = 0 if side in "nw" else tk.END
        self._add_widget_to_pane(new_pane, swidget, position=new_pos)
        self._raise_panes(new_pane)

    def _move_to_widget_side(
        self, swidget: DockWidget, twidget: DockWidget, side
    ):
        """Move 'swidget' to one side of 'twidget'."""
        logger.debug("Move to widget side: %s", side)
        if swidget == twidget:
            logger.debug("No move needed.")
            return

        tpane = twidget.parent_pane
        # move_in_same_pane = swidget.parent_pane == tpane
        simple_sides = "ns" if tpane.orient == tk.VERTICAL else "we"

        if side in simple_sides:
            index = self._get_position_in_pane(twidget)
            if side not in "nw":
                index += 1
                index = tk.END if index >= tpane.count else index
            self._move_into_pane_side(swidget, tpane, side, position=index)
        else:
            new_orient = (
                tk.HORIZONTAL if tpane.orient == tk.VERTICAL else tk.VERTICAL
            )
            self._replace_widget_with_pane(twidget, new_orient)
            self._move_into_pane_side(swidget, twidget.parent_pane, side)

    def _get_position_in_pane(self, widget):
        index = tk.END
        if isinstance(widget, DockWidget):
            pane = widget.parent_pane
            if pane:
                panes: tuple = pane.panedw.panes()
                index = panes.index(str(widget.noteb))
        elif isinstance(widget, DockPane):
            pane = widget.parent_pane
            if pane:
                panes: tuple = pane.panedw.panes()
                index = panes.index(str(widget))
        return index

    def _replace_widget_with_pane(self, widget: DockWidget, orient):
        """Creates new pane, move widget inside it and
        puts the pane in the current position of widget.
        """
        parent_pane = widget.parent_pane
        widget_pos = self._get_position_in_pane(widget)
        new_pane = self.new_pane(orient=orient)
        self._add_pane_to_pane(parent_pane, new_pane, position=widget_pos)
        side = "n" if orient == tk.VERTICAL else "w"
        self._move_into_pane_side(widget, new_pane, side, position=tk.END)

    def _raise_panes(self, widget):
        if isinstance(widget, DockPane):
            widget.tkraise()
            for w in widget.panedw.panes():
                self._raise_panes(widget.nametowidget(w))
        elif isinstance(widget, ttk.Notebook):
            widget.tkraise()
            for tab in widget.tabs():
                self._raise_panes(widget.nametowidget(tab))
        elif not isinstance(widget, tk.Canvas):
            widget.tkraise()

    def _clear_notebook(self, nb: ttk.Notebook):
        """Destroy notebook if does not have children."""
        if not nb.tabs():
            nb.destroy()

    def _clear_pane(self, pane: DockPane):
        """If pane does not have children, destroys it.
        Otherwise tries to simplificate it.
        """
        if pane.count == 0:
            pane.detach_from_parent()
            pane.destroy()
            return
        logger.debug("Simplify pane for: %s parent: %s", pane, pane.parent_pane)
        if pane.parent_pane:
            parent = pane.parent_pane
            if parent.orient == pane.orient:
                # FIXME: Pane can be simplified, try moving content to parent pane.?
                logger.debug(
                    "Pane can be simplified, try moving content to parent pane."
                )
            else:
                logger.debug(
                    "Pane orients differ: %s %s", parent.orient, pane.orient
                )
                if pane.count == 1:
                    logger.debug(
                        "Pane can be simplified, pane has only one child."
                    )
                    widget = pane.nametowidget(pane.panedw.panes()[0])
                    if isinstance(widget, ttk.Notebook):
                        tabs = widget.tabs()
                        if len(tabs) > 1:
                            # FIXME: simplify a pane with grouped widgets.
                            logger.warn("Grouped widgets not managed yet.")
                            logger.warn("Aborting simplification.")
                            return
                        widget = pane.nametowidget(widget.tabs()[0])
                    side = "n" if parent.orient == tk.VERTICAL else "w"
                    pos = self._get_position_in_pane(pane)
                    logger.debug("Trying to move: %s", widget)
                    if isinstance(widget, DockPane):
                        self._add_pane_to_pane(parent, widget, position=pos)
                    else:
                        self._move_into_pane_side(
                            widget, parent, side, position=pos
                        )

    def save_layout(self):
        """Return a dictionary with information of current layout."""
        config = {}
        if self.main_pane:
            layout = {}
            self._create_layout(self.main_pane, layout)
            config = {
                "ver": 1,
                "main": self.main_pane.uid,
                "layout": layout,
                "dwl": [dw_uid for dw_uid in self.dock_widgets],
            }
        return config

    def load_layout(self, config: dict):
        """Restore layout from previusly saved config dictionary."""
        if not self._layout_config_valid(config):
            raise ValueError("Invalid layout configuration.")
        self._remove_layout()
        main_uid = config["main"]
        pane = self._build_layout(main_uid, config["layout"])
        self.set_main_pane(pane)

    def _emit_layout_changed(self):
        """Schedule a notify of layout changed."""
        if self._emit_cbid is not None:
            self.after_cancel(self._emit_cbid)
        self._emit_cbid = self.after(800, self._notify_layout_changed)

    def _notify_layout_changed(self):
        self.event_generate(self.EVENT_LAYOUT_CHANGED)
        self._emit_cbid = None

    def _layout_config_valid(self, config):
        valid = True
        # validate "version"
        if config.get("ver", 0) != 1:
            valid = False
        # validate widgets exists
        if valid and "dwl" in config:
            for uid in config["dwl"]:
                if uid not in self.dock_widgets:
                    valid = False
                    break
        else:
            valid = False

        return valid

    def _create_layout(self, pane: DockPane, layout: dict):
        conf = {"orient": pane.orient}
        children = []
        layout[pane.uid] = {
            "cnf": conf,
            "ch": children,
        }
        # if len(pane.panedw.panes()) > 1:
        #    print(">>>", pane.uid),
        #    print(">>>", pane.panedw.sashpos(0))
        for child in pane.sorted_dw_children:
            if isinstance(child, DockPane):
                children.append(child.uid)
                self._create_layout(child, layout)
            else:
                dw = {
                    "uid": child.uid,
                    "grouped": child.is_grouped,
                    "weight": int(child.dw_options.get("weight", 1)),
                }
                children.append(dw)

    def _build_layout(self, pane_uid, layout):
        pane = self.new_pane(uid=pane_uid, **layout[pane_uid]["cnf"])
        for widget_or_pane in layout[pane_uid]["ch"]:
            if isinstance(widget_or_pane, dict):
                child_uid = widget_or_pane["uid"]
                grouped = widget_or_pane["grouped"]
                weight = widget_or_pane["weight"]
                if child_uid in self.dock_widgets:
                    pane.add_widget(
                        self.dock_widgets[child_uid],
                        grouped=grouped,
                        weight=weight,
                    )
            else:
                child_pane = self._build_layout(widget_or_pane, layout)
                pane.add_pane(child_pane)
        return pane

    def _remove_layout(self):
        """Remove all panes and set main pane to None"""

        def _recursive_remove(pane):
            for child in pane.dw_children:
                if isinstance(child, DockPane):
                    _recursive_remove(child)
                else:
                    child.detach_from_parent()
            pane.detach_from_parent()
            pane.destroy()

        if self.main_pane:
            _recursive_remove(self.main_pane)
            self._main_pane = None



# encoding: utf-8
from pygubu.api.v1 import (
    BuilderObject,
    register_widget,
    register_custom_property,
)
from pygubu.plugins.ttk.ttkstdwidgets import TTKFrame
from pygubu.i18n import _
from pygubu.widgets.calendarframe import CalendarFrame


class CalendarFrameBuilder(BuilderObject):
    class_ = CalendarFrame
    OPTIONS_STANDARD = TTKFrame.OPTIONS_STANDARD
    OPTIONS_SPECIFIC = TTKFrame.OPTIONS_SPECIFIC
    OPTIONS_CUSTOM = (
        "firstweekday",
        "year",
        "month",
        "calendarfg",
        "calendarbg",
        "headerfg",
        "headerbg",
        "selectbg",
        "selectfg",
        "state",
        "markbg",
        "markfg",
        "linewidth",
    )
    ro_properties = TTKFrame.ro_properties
    properties = OPTIONS_STANDARD + OPTIONS_SPECIFIC + OPTIONS_CUSTOM
    virtual_events = ("<<CalendarFrameDateSelected>>",)


_builder_id = "pygubu.builder.widgets.calendarframe"
register_widget(
    _builder_id,
    CalendarFrameBuilder,
    "CalendarFrame",
    ("ttk", _("Pygubu Widgets")),
)

register_custom_property(
    _builder_id,
    "state",
    "choice",
    values=("", "normal", "disabled"),
    state="readonly",
)
register_custom_property(
    _builder_id,
    "firstweekday",
    "choice",
    values=("0", "6"),
    state="readonly",
    default_value="6",
)
register_custom_property(_builder_id, "year", "entry")
register_custom_property(
    _builder_id,
    "month",
    "choice",
    values=("", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"),
    state="readonly",
)
register_custom_property(
    _builder_id,
    "linewidth",
    "choice",
    values=("1", "2", "3", "4"),
    state="readonly",
    default_value="1",
)
register_custom_property(_builder_id, "calendarfg", "colorentry")
register_custom_property(_builder_id, "calendarbg", "colorentry")
register_custom_property(_builder_id, "headerfg", "colorentry")
register_custom_property(_builder_id, "headerbg", "colorentry")
register_custom_property(_builder_id, "selectbg", "colorentry")
register_custom_property(_builder_id, "selectfg", "colorentry")
register_custom_property(_builder_id, "markbg", "colorentry")
register_custom_property(_builder_id, "markfg", "colorentry")
